from collections import namedtuple

from veros.variables import Variable
from veros.settings import Setting

VerosPlugin = namedtuple(
    "VerosPlugin",
    [
        "name",
        "module",
        "setup_entrypoint",
        "run_entrypoint",
        "settings",
        "variables",
        "dimensions",
        "diagnostics",
    ],
)


def load_plugin(module):
    from veros.diagnostics.base import VerosDiagnostic

    modname = module.__name__

    if not hasattr(module, "__VEROS_INTERFACE__"):
        raise RuntimeError(f"module {modname} is not a valid Veros plugin")

    interface = module.__VEROS_INTERFACE__

    setup_entrypoint = interface.get("setup_entrypoint")

    if not callable(setup_entrypoint):
        raise RuntimeError(f"module {modname} is missing a valid setup entrypoint")

    run_entrypoint = interface.get("run_entrypoint")

    if not callable(run_entrypoint):
        raise RuntimeError(f"module {modname} is missing a valid run entrypoint")

    name = interface.get("name", modname)

    settings = interface.get("settings", {})
    for setting, val in settings.items():
        if not isinstance(val, Setting):
            raise TypeError(f"got unexpected type {type(val)} for setting {setting}")

    variables = interface.get("variables", {})
    for variable, val in variables.items():
        if not isinstance(val, Variable):
            raise TypeError(f"got unexpected type {type(val)} for variable {variable}")

    dimensions = interface.get("dimensions", {})
    for dim, val in dimensions.items():
        if not isinstance(val, (str, int)):
            raise TypeError(f"got unexpected type {type(val)} for dimension {dim}")

    diagnostics = interface.get("diagnostics", [])
    for diagnostic in diagnostics:
        if not issubclass(diagnostic, VerosDiagnostic):
            raise TypeError(f"got unexpected type {type(diagnostic)} for diagnostic {diagnostic}")

    return VerosPlugin(
        name=name,
        module=module,
        setup_entrypoint=setup_entrypoint,
        run_entrypoint=run_entrypoint,
        settings=settings,
        variables=variables,
        dimensions=dimensions,
        diagnostics=diagnostics,
    )



from veros.tools.assets import get_assets  # noqa: F401
from veros.tools.setup import (  # noqa: F401
    interpolate,
    fill_holes,
    get_periodic_interval,
    make_cyclic,
    get_coastline_distance,
    get_uniform_grid_steps,
    get_stretched_grid_steps,
    get_vinokur_grid_steps,
)



"""
solve two dimensional Possion equation
     A * dpsi = forc,  where A = nabla_h^2
with Dirichlet boundary conditions
used for streamfunction
"""


from veros import veros_kernel, veros_routine, KernelOutput
from veros.variables import allocate
from veros.core import utilities as mainutils
from veros.core.operators import update, update_add, at, for_loop
from veros.core.operators import numpy as npx
from veros.core.external import line_integrals
from veros.core.external.solvers import get_linear_solver


@veros_routine
def solve_streamfunction(state):
    vs = state.variables

    state_update, (forc, uloc, vloc) = prepare_forcing(state)
    vs.update(state_update)

    linear_solver = get_linear_solver(state)
    linear_sol = linear_solver.solve(state, forc, vs.dpsi[..., vs.taup1])
    vs.dpsi = update(vs.dpsi, at[..., vs.taup1], linear_sol)

    vs.update(barotropic_velocity_update(state, uloc=uloc, vloc=vloc))


@veros_kernel
def prepare_forcing(state):
    vs = state.variables
    settings = state.settings

    # hydrostatic pressure
    vs.p_hydro = update(
        vs.p_hydro,
        at[:, :, -1],
        0.5 * vs.rho[:, :, -1, vs.tau] * settings.grav / settings.rho_0 * vs.dzw[-1] * vs.maskT[:, :, -1],
    )

    def compute_p_hydro(k_inv, p_hydro):
        k = settings.nz - k_inv - 1
        p_hydro = update(
            p_hydro,
            at[..., k],
            vs.maskT[:, :, k]
            * (
                p_hydro[:, :, k + 1]
                + 0.5
                * vs.dzw[k]
                * settings.grav
                / settings.rho_0
                * (vs.rho[:, :, k + 1, vs.tau] + vs.rho[:, :, k, vs.tau])
            ),
        )
        return p_hydro

    vs.p_hydro = for_loop(1, settings.nz, compute_p_hydro, vs.p_hydro)

    # add hydrostatic pressure gradient
    vs.du = update_add(
        vs.du,
        at[2:-2, 2:-2, :, vs.tau],
        -(vs.p_hydro[3:-1, 2:-2, :] - vs.p_hydro[2:-2, 2:-2, :])
        / (vs.cost[npx.newaxis, 2:-2, npx.newaxis] * vs.dxu[2:-2, npx.newaxis, npx.newaxis])
        * vs.maskU[2:-2, 2:-2, :],
    )
    vs.dv = update_add(
        vs.dv,
        at[2:-2, 2:-2, :, vs.tau],
        -(vs.p_hydro[2:-2, 3:-1, :] - vs.p_hydro[2:-2, 2:-2, :])
        / vs.dyu[npx.newaxis, 2:-2, npx.newaxis]
        * vs.maskV[2:-2, 2:-2, :],
    )

    # forcing for barotropic streamfunction
    uloc = npx.sum((vs.du[:, :, :, vs.tau] + vs.du_mix) * vs.maskU * vs.dzt, axis=(2,)) * vs.hur
    vloc = npx.sum((vs.dv[:, :, :, vs.tau] + vs.dv_mix) * vs.maskV * vs.dzt, axis=(2,)) * vs.hvr

    uloc = mainutils.enforce_boundaries(uloc, settings.enable_cyclic_x)
    vloc = mainutils.enforce_boundaries(vloc, settings.enable_cyclic_x)

    forc = allocate(state.dimensions, ("xt", "yt"))
    forc = update(
        forc,
        at[2:-2, 2:-2],
        (vloc[3:-1, 2:-2] - vloc[2:-2, 2:-2]) / (vs.cosu[2:-2] * vs.dxu[2:-2, npx.newaxis])
        - (vs.cost[3:-1] * uloc[2:-2, 3:-1] - vs.cost[2:-2] * uloc[2:-2, 2:-2]) / (vs.cosu[2:-2] * vs.dyu[2:-2]),
    )

    # solve for interior streamfunction
    vs.dpsi = update(vs.dpsi, at[:, :, vs.taup1], 2 * vs.dpsi[:, :, vs.tau] - vs.dpsi[:, :, vs.taum1])

    return KernelOutput(du=vs.du, dv=vs.dv, dpsi=vs.dpsi, p_hydro=vs.p_hydro), (forc, uloc, vloc)


@veros_kernel
def barotropic_velocity_update(state, uloc, vloc):
    """
    solve for barotropic streamfunction
    """
    vs = state.variables
    settings = state.settings

    vs.dpsi = update(
        vs.dpsi, at[:, :, vs.taup1], mainutils.enforce_boundaries(vs.dpsi[:, :, vs.taup1], settings.enable_cyclic_x)
    )

    line_forc = allocate(state.dimensions, ("isle",))

    if state.dimensions["isle"] > 1:
        # calculate island integrals of forcing, keep psi constant on island 1
        line_forc = update(
            line_forc,
            at[1:],
            line_integrals.line_integrals(state, uloc=uloc[..., npx.newaxis], vloc=vloc[..., npx.newaxis], kind="same")[
                1:
            ],
        )

        # calculate island integrals of interior streamfunction
        uloc = update(uloc, at[...], 0.0)
        vloc = update(vloc, at[...], 0.0)
        uloc = update(
            uloc,
            at[1:, 1:],
            -1
            * vs.maskU[1:, 1:, -1]
            * (vs.dpsi[1:, 1:, vs.taup1] - vs.dpsi[1:, :-1, vs.taup1])
            / vs.dyt[npx.newaxis, 1:]
            * vs.hur[1:, 1:],
        )
        vloc = update(
            vloc,
            at[1:, 1:],
            vs.maskV[1:, 1:, -1]
            * (vs.dpsi[1:, 1:, vs.taup1] - vs.dpsi[:-1, 1:, vs.taup1])
            / (vs.cosu[npx.newaxis, 1:] * vs.dxt[1:, npx.newaxis])
            * vs.hvr[1:, 1:],
        )
        line_forc = update_add(
            line_forc,
            at[1:],
            -line_integrals.line_integrals(
                state, uloc=uloc[..., npx.newaxis], vloc=vloc[..., npx.newaxis], kind="same"
            )[1:],
        )

        # solve for time dependent boundary values
        vs.dpsin = update(vs.dpsin, at[1:, vs.tau], npx.linalg.solve(vs.line_psin[1:, 1:], line_forc[1:]))

    # integrate barotropic and baroclinic velocity forward in time
    vs.psi = update(
        vs.psi,
        at[:, :, vs.taup1],
        vs.psi[:, :, vs.tau]
        + settings.dt_mom
        * ((1.5 + settings.AB_eps) * vs.dpsi[:, :, vs.taup1] - (0.5 + settings.AB_eps) * vs.dpsi[:, :, vs.tau]),
    )
    vs.psi = update_add(
        vs.psi,
        at[:, :, vs.taup1],
        settings.dt_mom
        * npx.sum(
            ((1.5 + settings.AB_eps) * vs.dpsin[1:, vs.tau] - (0.5 + settings.AB_eps) * vs.dpsin[1:, vs.taum1])
            * vs.psin[:, :, 1:],
            axis=2,
        ),
    )
    vs.u = update(
        vs.u,
        at[:, :, :, vs.taup1],
        vs.u[:, :, :, vs.tau]
        + settings.dt_mom
        * (
            vs.du_mix
            + (1.5 + settings.AB_eps) * vs.du[:, :, :, vs.tau]
            - (0.5 + settings.AB_eps) * vs.du[:, :, :, vs.taum1]
        )
        * vs.maskU,
    )
    vs.v = update(
        vs.v,
        at[:, :, :, vs.taup1],
        vs.v[:, :, :, vs.tau]
        + settings.dt_mom
        * (
            vs.dv_mix
            + (1.5 + settings.AB_eps) * vs.dv[:, :, :, vs.tau]
            - (0.5 + settings.AB_eps) * vs.dv[:, :, :, vs.taum1]
        )
        * vs.maskV,
    )

    # subtract incorrect vertical mean from baroclinic velocity
    uloc = npx.sum(vs.u[:, :, :, vs.taup1] * vs.maskU * vs.dzt, axis=2)
    vloc = npx.sum(vs.v[:, :, :, vs.taup1] * vs.maskV * vs.dzt, axis=2)
    vs.u = update_add(vs.u, at[:, :, :, vs.taup1], -uloc[:, :, npx.newaxis] * vs.maskU * vs.hur[:, :, npx.newaxis])
    vs.v = update_add(vs.v, at[:, :, :, vs.taup1], -vloc[:, :, npx.newaxis] * vs.maskV * vs.hvr[:, :, npx.newaxis])

    # add barotropic mode to baroclinic velocity
    vs.u = update_add(
        vs.u,
        at[2:-2, 2:-2, :, vs.taup1],
        -1
        * vs.maskU[2:-2, 2:-2, :]
        * (vs.psi[2:-2, 2:-2, vs.taup1, npx.newaxis] - vs.psi[2:-2, 1:-3, vs.taup1, npx.newaxis])
        / vs.dyt[npx.newaxis, 2:-2, npx.newaxis]
        * vs.hur[2:-2, 2:-2, npx.newaxis],
    )
    vs.v = update_add(
        vs.v,
        at[2:-2, 2:-2, :, vs.taup1],
        vs.maskV[2:-2, 2:-2, :]
        * (vs.psi[2:-2, 2:-2, vs.taup1, npx.newaxis] - vs.psi[1:-3, 2:-2, vs.taup1, npx.newaxis])
        / (vs.cosu[2:-2, npx.newaxis] * vs.dxt[2:-2, npx.newaxis, npx.newaxis])
        * vs.hvr[2:-2, 2:-2][:, :, npx.newaxis],
    )

    return KernelOutput(u=vs.u, v=vs.v, psi=vs.psi, dpsi=vs.dpsi, dpsin=vs.dpsin)



YEAR_LENGTH = 360.0
X_TO_SECONDS = {
    "seconds": 1.0,
    "minutes": 60.0,
    "hours": 60.0 * 60.0,
    "days": 24.0 * 60.0 * 60.0,
    "years": YEAR_LENGTH * 24.0 * 60.0 * 60.0,
}
SECONDS_TO_X = {key: 1.0 / val for key, val in X_TO_SECONDS.items()}


def convert_time(time_value, in_unit, out_unit):
    return time_value * X_TO_SECONDS[in_unit] * SECONDS_TO_X[out_unit]


def format_time(time_value, in_unit="seconds"):
    all_units = X_TO_SECONDS.keys()
    val_in_all_units = {u: convert_time(time_value, in_unit, u) for u in all_units}
    valid_units = {u: v for u, v in val_in_all_units.items() if v >= 1.0}
    if valid_units:
        best_unit = min(valid_units, key=valid_units.get)
    else:
        best_unit = "seconds"
    return val_in_all_units[best_unit], best_unit



from veros.core.operators import numpy as npx

from veros import veros_routine, veros_kernel, KernelOutput
from veros.variables import allocate
from veros.core import numerics, utilities, isoneutral
from veros.core.operators import update, update_add, at


@veros_kernel
def explicit_vert_friction(state):
    """
    explicit vertical friction
    dissipation is calculated and added to K_diss_v
    """
    vs = state.variables

    diss = allocate(state.dimensions, ("xt", "yu", "zt"))
    flux_top = allocate(state.dimensions, ("xt", "yu", "zt"))

    """
    vertical friction of zonal momentum
    """
    fxa = 0.5 * (vs.kappaM[1:-2, 1:-2, :-1] + vs.kappaM[2:-1, 1:-2, :-1])
    flux_top = update(
        flux_top,
        at[1:-2, 1:-2, :-1],
        fxa
        * (vs.u[1:-2, 1:-2, 1:, vs.tau] - vs.u[1:-2, 1:-2, :-1, vs.tau])
        / vs.dzw[npx.newaxis, npx.newaxis, :-1]
        * vs.maskU[1:-2, 1:-2, 1:]
        * vs.maskU[1:-2, 1:-2, :-1],
    )
    flux_top = update(flux_top, at[:, :, -1], 0.0)
    vs.du_mix = update(vs.du_mix, at[:, :, 0], flux_top[:, :, 0] / vs.dzt[0] * vs.maskU[:, :, 0])
    vs.du_mix = update(
        vs.du_mix, at[:, :, 1:], (flux_top[:, :, 1:] - flux_top[:, :, :-1]) / vs.dzt[1:] * vs.maskU[:, :, 1:]
    )

    """
    diagnose dissipation by vertical friction of zonal momentum
    """
    diss = update(
        diss,
        at[1:-2, 1:-2, :-1],
        (vs.u[1:-2, 1:-2, 1:, vs.tau] - vs.u[1:-2, 1:-2, :-1, vs.tau])
        * flux_top[1:-2, 1:-2, :-1]
        / vs.dzw[npx.newaxis, npx.newaxis, :-1],
    )
    diss = update(diss, at[:, :, -1], 0.0)
    diss = numerics.ugrid_to_tgrid(state, diss)
    vs.K_diss_v = vs.K_diss_v + diss

    """
    vertical friction of meridional momentum
    """
    fxa = 0.5 * (vs.kappaM[1:-2, 1:-2, :-1] + vs.kappaM[1:-2, 2:-1, :-1])
    flux_top = update(
        flux_top,
        at[1:-2, 1:-2, :-1],
        fxa
        * (vs.v[1:-2, 1:-2, 1:, vs.tau] - vs.v[1:-2, 1:-2, :-1, vs.tau])
        / vs.dzw[npx.newaxis, npx.newaxis, :-1]
        * vs.maskV[1:-2, 1:-2, 1:]
        * vs.maskV[1:-2, 1:-2, :-1],
    )
    flux_top = update(flux_top, at[:, :, -1], 0.0)
    vs.dv_mix = update(
        vs.dv_mix,
        at[:, :, 1:],
        (flux_top[:, :, 1:] - flux_top[:, :, :-1]) / vs.dzt[npx.newaxis, npx.newaxis, 1:] * vs.maskV[:, :, 1:],
    )
    vs.dv_mix = update(vs.dv_mix, at[:, :, 0], flux_top[:, :, 0] / vs.dzt[0] * vs.maskV[:, :, 0])

    """
    diagnose dissipation by vertical friction of meridional momentum
    """
    diss = update(
        diss,
        at[1:-2, 1:-2, :-1],
        (vs.v[1:-2, 1:-2, 1:, vs.tau] - vs.v[1:-2, 1:-2, :-1, vs.tau])
        * flux_top[1:-2, 1:-2, :-1]
        / vs.dzw[npx.newaxis, npx.newaxis, :-1],
    )
    diss = update(diss, at[:, :, -1], 0.0)
    diss = numerics.vgrid_to_tgrid(state, diss)
    vs.K_diss_v = vs.K_diss_v + diss

    return KernelOutput(du_mix=vs.du_mix, dv_mix=vs.dv_mix, K_diss_v=vs.K_diss_v)


@veros_kernel
def implicit_vert_friction(state):
    """
    vertical friction
    dissipation is calculated and added to K_diss_v
    """
    vs = state.variables
    settings = state.settings

    diss = allocate(state.dimensions, ("xt", "yu", "zt"))
    a_tri = allocate(state.dimensions, ("xt", "yu", "zt"))[1:-2, 1:-2]
    b_tri = allocate(state.dimensions, ("xt", "yu", "zt"))[1:-2, 1:-2]
    c_tri = allocate(state.dimensions, ("xt", "yu", "zt"))[1:-2, 1:-2]
    d_tri = allocate(state.dimensions, ("xt", "yu", "zt"))[1:-2, 1:-2]
    delta = allocate(state.dimensions, ("xt", "yu", "zt"))[1:-2, 1:-2]
    flux_top = allocate(state.dimensions, ("xt", "yu", "zt"))

    """
    implicit vertical friction of zonal momentum
    """
    kss = npx.maximum(vs.kbot[1:-2, 1:-2], vs.kbot[2:-1, 1:-2])
    _, water_mask, edge_mask = utilities.create_water_masks(kss, settings.nz)

    fxa = 0.5 * (vs.kappaM[1:-2, 1:-2, :-1] + vs.kappaM[2:-1, 1:-2, :-1])
    delta = update(
        delta, at[:, :, :-1], settings.dt_mom / vs.dzw[:-1] * fxa * vs.maskU[1:-2, 1:-2, 1:] * vs.maskU[1:-2, 1:-2, :-1]
    )
    a_tri = update(a_tri, at[:, :, 1:], -delta[:, :, :-1] / vs.dzt[npx.newaxis, npx.newaxis, 1:])
    b_tri = update(b_tri, at[:, :, 1:], 1 + delta[:, :, :-1] / vs.dzt[npx.newaxis, npx.newaxis, 1:])
    b_tri = update_add(b_tri, at[:, :, 1:-1], delta[:, :, 1:-1] / vs.dzt[npx.newaxis, npx.newaxis, 1:-1])
    b_tri_edge = 1 + delta / vs.dzt[npx.newaxis, npx.newaxis, :]
    c_tri = update(c_tri, at[...], -delta / vs.dzt[npx.newaxis, npx.newaxis, :])
    d_tri = update(d_tri, at[...], vs.u[1:-2, 1:-2, :, vs.tau])

    res = utilities.solve_implicit(a_tri, b_tri, c_tri, d_tri, water_mask, b_edge=b_tri_edge, edge_mask=edge_mask)
    vs.u = update(vs.u, at[1:-2, 1:-2, :, vs.taup1], npx.where(water_mask, res, vs.u[1:-2, 1:-2, :, vs.taup1]))
    vs.du_mix = update(
        vs.du_mix, at[1:-2, 1:-2], (vs.u[1:-2, 1:-2, :, vs.taup1] - vs.u[1:-2, 1:-2, :, vs.tau]) / settings.dt_mom
    )

    """
    diagnose dissipation by vertical friction of zonal momentum
    """
    fxa = 0.5 * (vs.kappaM[1:-2, 1:-2, :-1] + vs.kappaM[2:-1, 1:-2, :-1])
    flux_top = update(
        flux_top,
        at[1:-2, 1:-2, :-1],
        fxa
        * (vs.u[1:-2, 1:-2, 1:, vs.taup1] - vs.u[1:-2, 1:-2, :-1, vs.taup1])
        / vs.dzw[:-1]
        * vs.maskU[1:-2, 1:-2, 1:]
        * vs.maskU[1:-2, 1:-2, :-1],
    )
    diss = update(
        diss,
        at[1:-2, 1:-2, :-1],
        (vs.u[1:-2, 1:-2, 1:, vs.tau] - vs.u[1:-2, 1:-2, :-1, vs.tau]) * flux_top[1:-2, 1:-2, :-1] / vs.dzw[:-1],
    )
    diss = update(diss, at[:, :, -1], 0.0)
    diss = numerics.ugrid_to_tgrid(state, diss)
    vs.K_diss_v = vs.K_diss_v + diss

    """
    implicit vertical friction of meridional momentum
    """
    kss = npx.maximum(vs.kbot[1:-2, 1:-2], vs.kbot[1:-2, 2:-1])
    _, water_mask, edge_mask = utilities.create_water_masks(kss, settings.nz)

    fxa = 0.5 * (vs.kappaM[1:-2, 1:-2, :-1] + vs.kappaM[1:-2, 2:-1, :-1])
    delta = update(
        delta,
        at[:, :, :-1],
        settings.dt_mom
        / vs.dzw[npx.newaxis, npx.newaxis, :-1]
        * fxa
        * vs.maskV[1:-2, 1:-2, 1:]
        * vs.maskV[1:-2, 1:-2, :-1],
    )
    a_tri = update(a_tri, at[:, :, 1:], -delta[:, :, :-1] / vs.dzt[npx.newaxis, npx.newaxis, 1:])
    b_tri = update(b_tri, at[:, :, 1:], 1 + delta[:, :, :-1] / vs.dzt[npx.newaxis, npx.newaxis, 1:])
    b_tri = update_add(b_tri, at[:, :, 1:-1], delta[:, :, 1:-1] / vs.dzt[npx.newaxis, npx.newaxis, 1:-1])
    b_tri_edge = 1 + delta / vs.dzt[npx.newaxis, npx.newaxis, :]
    c_tri = update(c_tri, at[:, :, :-1], -delta[:, :, :-1] / vs.dzt[npx.newaxis, npx.newaxis, :-1])
    c_tri = update(c_tri, at[:, :, -1], 0.0)
    d_tri = update(d_tri, at[...], vs.v[1:-2, 1:-2, :, vs.tau])

    res = utilities.solve_implicit(a_tri, b_tri, c_tri, d_tri, water_mask, b_edge=b_tri_edge, edge_mask=edge_mask)
    vs.v = update(vs.v, at[1:-2, 1:-2, :, vs.taup1], npx.where(water_mask, res, vs.v[1:-2, 1:-2, :, vs.taup1]))
    vs.dv_mix = update(
        vs.dv_mix, at[1:-2, 1:-2], (vs.v[1:-2, 1:-2, :, vs.taup1] - vs.v[1:-2, 1:-2, :, vs.tau]) / settings.dt_mom
    )

    """
    diagnose dissipation by vertical friction of meridional momentum
    """
    fxa = 0.5 * (vs.kappaM[1:-2, 1:-2, :-1] + vs.kappaM[1:-2, 2:-1, :-1])
    flux_top = update(
        flux_top,
        at[1:-2, 1:-2, :-1],
        fxa
        * (vs.v[1:-2, 1:-2, 1:, vs.taup1] - vs.v[1:-2, 1:-2, :-1, vs.taup1])
        / vs.dzw[:-1]
        * vs.maskV[1:-2, 1:-2, 1:]
        * vs.maskV[1:-2, 1:-2, :-1],
    )
    diss = update(
        diss,
        at[1:-2, 1:-2, :-1],
        (vs.v[1:-2, 1:-2, 1:, vs.tau] - vs.v[1:-2, 1:-2, :-1, vs.tau]) * flux_top[1:-2, 1:-2, :-1] / vs.dzw[:-1],
    )
    diss = update(diss, at[:, :, -1], 0.0)
    diss = numerics.vgrid_to_tgrid(state, diss)
    vs.K_diss_v = vs.K_diss_v + diss

    return KernelOutput(u=vs.u, v=vs.v, du_mix=vs.du_mix, dv_mix=vs.dv_mix, K_diss_v=vs.K_diss_v)


@veros_kernel
def rayleigh_friction(state):
    """
    interior Rayleigh friction
    dissipation is calculated and added to K_diss_bot
    """
    vs = state.variables
    settings = state.settings

    vs.du_mix = update_add(vs.du_mix, at[...], -1 * vs.maskU * settings.r_ray * vs.u[..., vs.tau])
    if settings.enable_conserve_energy:
        diss = vs.maskU * settings.r_ray * vs.u[..., vs.tau] ** 2
        vs.K_diss_bot = update_add(vs.K_diss_bot, at[...], numerics.calc_diss_u(state, diss))

    vs.dv_mix = update_add(vs.dv_mix, at[...], -1 * vs.maskV * settings.r_ray * vs.v[..., vs.tau])
    if settings.enable_conserve_energy:
        diss = vs.maskV * settings.r_ray * vs.v[..., vs.tau] ** 2
        vs.K_diss_bot = update_add(vs.K_diss_bot, at[...], numerics.calc_diss_v(state, diss))

    return KernelOutput(du_mix=vs.du_mix, dv_mix=vs.dv_mix, K_diss_bot=vs.K_diss_bot)


@veros_kernel
def linear_bottom_friction(state):
    """
    linear bottom friction
    dissipation is calculated and added to K_diss_bot
    """
    vs = state.variables
    settings = state.settings

    if settings.enable_bottom_friction_var:
        """
        with spatially varying coefficient
        """
        k = npx.maximum(vs.kbot[1:-2, 2:-2], vs.kbot[2:-1, 2:-2]) - 1
        mask = npx.arange(settings.nz) == k[:, :, npx.newaxis]
        vs.du_mix = update_add(
            vs.du_mix,
            at[1:-2, 2:-2],
            -(vs.maskU[1:-2, 2:-2] * vs.r_bot_var_u[1:-2, 2:-2, npx.newaxis]) * vs.u[1:-2, 2:-2, :, vs.tau] * mask,
        )
        if settings.enable_conserve_energy:
            diss = allocate(state.dimensions, ("xt", "yu", "zt"))
            diss = update(
                diss,
                at[1:-2, 2:-2],
                vs.maskU[1:-2, 2:-2]
                * vs.r_bot_var_u[1:-2, 2:-2, npx.newaxis]
                * vs.u[1:-2, 2:-2, :, vs.tau] ** 2
                * mask,
            )
            vs.K_diss_bot = update_add(vs.K_diss_bot, at[...], numerics.calc_diss_u(state, diss))

        k = npx.maximum(vs.kbot[2:-2, 2:-1], vs.kbot[2:-2, 1:-2]) - 1
        mask = npx.arange(settings.nz) == k[:, :, npx.newaxis]
        vs.dv_mix = update_add(
            vs.dv_mix,
            at[2:-2, 1:-2],
            -(vs.maskV[2:-2, 1:-2] * vs.r_bot_var_v[2:-2, 1:-2, npx.newaxis]) * vs.v[2:-2, 1:-2, :, vs.tau] * mask,
        )
        if settings.enable_conserve_energy:
            diss = allocate(state.dimensions, ("xt", "yu", "zt"))
            diss = update(
                diss,
                at[2:-2, 1:-2],
                vs.maskV[2:-2, 1:-2]
                * vs.r_bot_var_v[2:-2, 1:-2, npx.newaxis]
                * vs.v[2:-2, 1:-2, :, vs.tau] ** 2
                * mask,
            )
            vs.K_diss_bot = update_add(vs.K_diss_bot, at[...], numerics.calc_diss_v(state, diss))
    else:
        """
        with constant coefficient
        """
        k = npx.maximum(vs.kbot[1:-2, 2:-2], vs.kbot[2:-1, 2:-2]) - 1
        mask = npx.arange(settings.nz) == k[:, :, npx.newaxis]

        vs.du_mix = update_add(
            vs.du_mix, at[1:-2, 2:-2], -1 * vs.maskU[1:-2, 2:-2] * settings.r_bot * vs.u[1:-2, 2:-2, :, vs.tau] * mask
        )
        if settings.enable_conserve_energy:
            diss = allocate(state.dimensions, ("xt", "yu", "zt"))
            diss = update(
                diss, at[1:-2, 2:-2], vs.maskU[1:-2, 2:-2] * settings.r_bot * vs.u[1:-2, 2:-2, :, vs.tau] ** 2 * mask
            )
            vs.K_diss_bot = update_add(vs.K_diss_bot, at[...], numerics.calc_diss_u(state, diss))

        k = npx.maximum(vs.kbot[2:-2, 2:-1], vs.kbot[2:-2, 1:-2]) - 1
        mask = npx.arange(settings.nz) == k[:, :, npx.newaxis]

        vs.dv_mix = update_add(
            vs.dv_mix, at[2:-2, 1:-2], -1 * vs.maskV[2:-2, 1:-2] * settings.r_bot * vs.v[2:-2, 1:-2, :, vs.tau] * mask
        )
        if settings.enable_conserve_energy:
            diss = allocate(state.dimensions, ("xt", "yu", "zt"))
            diss = update(
                diss, at[2:-2, 1:-2], vs.maskV[2:-2, 1:-2] * settings.r_bot * vs.v[2:-2, 1:-2, :, vs.tau] ** 2 * mask
            )
            vs.K_diss_bot = update_add(vs.K_diss_bot, at[...], numerics.calc_diss_v(state, diss))

    return KernelOutput(du_mix=vs.du_mix, dv_mix=vs.dv_mix, K_diss_bot=vs.K_diss_bot)


@veros_kernel
def quadratic_bottom_friction(state):
    """
    quadratic bottom friction
    dissipation is calculated and added to K_diss_bot
    """
    vs = state.variables
    settings = state.settings

    # we might want to account for EKE in the drag, also a tidal residual
    k = npx.maximum(vs.kbot[1:-2, 2:-2], vs.kbot[2:-1, 2:-2]) - 1
    mask = k[..., npx.newaxis] == npx.arange(settings.nz)[npx.newaxis, npx.newaxis, :]
    fxa = (
        vs.maskV[1:-2, 2:-2, :] * vs.v[1:-2, 2:-2, :, vs.tau] ** 2
        + vs.maskV[1:-2, 1:-3, :] * vs.v[1:-2, 1:-3, :, vs.tau] ** 2
        + vs.maskV[2:-1, 2:-2, :] * vs.v[2:-1, 2:-2, :, vs.tau] ** 2
        + vs.maskV[2:-1, 1:-3, :] * vs.v[2:-1, 1:-3, :, vs.tau] ** 2
    )
    fxa = npx.sqrt(vs.u[1:-2, 2:-2, :, vs.tau] ** 2 + 0.25 * fxa)
    aloc = (
        vs.maskU[1:-2, 2:-2, :]
        * settings.r_quad_bot
        * vs.u[1:-2, 2:-2, :, vs.tau]
        * fxa
        / vs.dzt[npx.newaxis, npx.newaxis, :]
        * mask
    )
    vs.du_mix = update_add(vs.du_mix, at[1:-2, 2:-2, :], -aloc)

    if settings.enable_conserve_energy:
        diss = allocate(state.dimensions, ("xt", "yu", "zt"))
        diss = update(diss, at[1:-2, 2:-2, :], aloc * vs.u[1:-2, 2:-2, :, vs.tau])
        vs.K_diss_bot = update_add(vs.K_diss_bot, at[...], numerics.calc_diss_u(state, diss))

    k = npx.maximum(vs.kbot[2:-2, 1:-2], vs.kbot[2:-2, 2:-1]) - 1
    mask = k[..., npx.newaxis] == npx.arange(settings.nz)[npx.newaxis, npx.newaxis, :]
    fxa = (
        vs.maskU[2:-2, 1:-2, :] * vs.u[2:-2, 1:-2, :, vs.tau] ** 2
        + vs.maskU[1:-3, 1:-2, :] * vs.u[1:-3, 1:-2, :, vs.tau] ** 2
        + vs.maskU[2:-2, 2:-1, :] * vs.u[2:-2, 2:-1, :, vs.tau] ** 2
        + vs.maskU[1:-3, 2:-1, :] * vs.u[1:-3, 2:-1, :, vs.tau] ** 2
    )
    fxa = npx.sqrt(vs.v[2:-2, 1:-2, :, vs.tau] ** 2 + 0.25 * fxa)
    aloc = (
        vs.maskV[2:-2, 1:-2, :]
        * settings.r_quad_bot
        * vs.v[2:-2, 1:-2, :, vs.tau]
        * fxa
        / vs.dzt[npx.newaxis, npx.newaxis, :]
        * mask
    )
    vs.dv_mix = update_add(vs.dv_mix, at[2:-2, 1:-2, :], -aloc)

    if settings.enable_conserve_energy:
        diss = allocate(state.dimensions, ("xt", "yu", "zt"))
        diss = update(diss, at[2:-2, 1:-2, :], aloc * vs.v[2:-2, 1:-2, :, vs.tau])
        vs.K_diss_bot = update_add(vs.K_diss_bot, at[...], numerics.calc_diss_v(state, diss))

    return KernelOutput(du_mix=vs.du_mix, dv_mix=vs.dv_mix, K_diss_bot=vs.K_diss_bot)


@veros_kernel
def harmonic_friction(state):
    """
    horizontal harmonic friction
    dissipation is calculated and added to K_diss_h
    """
    vs = state.variables
    settings = state.settings

    diss = allocate(state.dimensions, ("xt", "yt", "zt"))
    flux_east = allocate(state.dimensions, ("xu", "yt", "zt"))
    flux_north = allocate(state.dimensions, ("xt", "yu", "zt"))

    """
    Zonal velocity
    """
    if settings.enable_hor_friction_cos_scaling:
        fxa = vs.cost**settings.hor_friction_cosPower
        flux_east = update(
            flux_east,
            at[:-1],
            settings.A_h
            * fxa[npx.newaxis, :, npx.newaxis]
            * (vs.u[1:, :, :, vs.tau] - vs.u[:-1, :, :, vs.tau])
            / (vs.cost * vs.dxt[1:, npx.newaxis])[:, :, npx.newaxis]
            * vs.maskU[1:]
            * vs.maskU[:-1],
        )
        fxa = vs.cosu**settings.hor_friction_cosPower
        flux_north = update(
            flux_north,
            at[:, :-1],
            settings.A_h
            * fxa[npx.newaxis, :-1, npx.newaxis]
            * (vs.u[:, 1:, :, vs.tau] - vs.u[:, :-1, :, vs.tau])
            / vs.dyu[npx.newaxis, :-1, npx.newaxis]
            * vs.maskU[:, 1:]
            * vs.maskU[:, :-1]
            * vs.cosu[npx.newaxis, :-1, npx.newaxis],
        )
        if settings.enable_noslip_lateral:
            flux_north = update_add(
                flux_north,
                at[:, :-1],
                2
                * settings.A_h
                * fxa[npx.newaxis, :-1, npx.newaxis]
                * (vs.u[:, 1:, :, vs.tau])
                / vs.dyu[npx.newaxis, :-1, npx.newaxis]
                * vs.maskU[:, 1:]
                * (1 - vs.maskU[:, :-1])
                * vs.cosu[npx.newaxis, :-1, npx.newaxis]
                - 2
                * settings.A_h
                * fxa[npx.newaxis, :-1, npx.newaxis]
                * (vs.u[:, :-1, :, vs.tau])
                / vs.dyu[npx.newaxis, :-1, npx.newaxis]
                * (1 - vs.maskU[:, 1:])
                * vs.maskU[:, :-1]
                * vs.cosu[npx.newaxis, :-1, npx.newaxis],
            )
    else:
        flux_east = update(
            flux_east,
            at[:-1, :, :],
            settings.A_h
            * (vs.u[1:, :, :, vs.tau] - vs.u[:-1, :, :, vs.tau])
            / (vs.cost * vs.dxt[1:, npx.newaxis])[:, :, npx.newaxis]
            * vs.maskU[1:]
            * vs.maskU[:-1],
        )
        flux_north = update(
            flux_north,
            at[:, :-1, :],
            settings.A_h
            * (vs.u[:, 1:, :, vs.tau] - vs.u[:, :-1, :, vs.tau])
            / vs.dyu[npx.newaxis, :-1, npx.newaxis]
            * vs.maskU[:, 1:]
            * vs.maskU[:, :-1]
            * vs.cosu[npx.newaxis, :-1, npx.newaxis],
        )
        if settings.enable_noslip_lateral:
            flux_north = update_add(
                flux_north,
                at[:, :-1],
                2
                * settings.A_h
                * vs.u[:, 1:, :, vs.tau]
                / vs.dyu[npx.newaxis, :-1, npx.newaxis]
                * vs.maskU[:, 1:]
                * (1 - vs.maskU[:, :-1])
                * vs.cosu[npx.newaxis, :-1, npx.newaxis]
                - 2
                * settings.A_h
                * vs.u[:, :-1, :, vs.tau]
                / vs.dyu[npx.newaxis, :-1, npx.newaxis]
                * (1 - vs.maskU[:, 1:])
                * vs.maskU[:, :-1]
                * vs.cosu[npx.newaxis, :-1, npx.newaxis],
            )

    flux_east = update(flux_east, at[-1, :, :], 0.0)
    flux_north = update(flux_north, at[:, -1, :], 0.0)

    """
    update tendency
    """
    vs.du_mix = update_add(
        vs.du_mix,
        at[2:-2, 2:-2, :],
        vs.maskU[2:-2, 2:-2]
        * (
            (flux_east[2:-2, 2:-2] - flux_east[1:-3, 2:-2])
            / (vs.cost[2:-2] * vs.dxu[2:-2, npx.newaxis])[:, :, npx.newaxis]
            + (flux_north[2:-2, 2:-2] - flux_north[2:-2, 1:-3])
            / (vs.cost[2:-2] * vs.dyt[2:-2])[npx.newaxis, :, npx.newaxis]
        ),
    )

    if settings.enable_conserve_energy:
        """
        diagnose dissipation by lateral friction
        """
        diss = update(
            diss,
            at[1:-2, 2:-2],
            0.5
            * (
                (vs.u[2:-1, 2:-2, :, vs.tau] - vs.u[1:-2, 2:-2, :, vs.tau]) * flux_east[1:-2, 2:-2]
                + (vs.u[1:-2, 2:-2, :, vs.tau] - vs.u[:-3, 2:-2, :, vs.tau]) * flux_east[:-3, 2:-2]
            )
            / (vs.cost[2:-2] * vs.dxu[1:-2, npx.newaxis])[:, :, npx.newaxis]
            + 0.5
            * (
                (vs.u[1:-2, 3:-1, :, vs.tau] - vs.u[1:-2, 2:-2, :, vs.tau]) * flux_north[1:-2, 2:-2]
                + (vs.u[1:-2, 2:-2, :, vs.tau] - vs.u[1:-2, 1:-3, :, vs.tau]) * flux_north[1:-2, 1:-3]
            )
            / (vs.cost[2:-2] * vs.dyt[2:-2])[npx.newaxis, :, npx.newaxis],
        )
        vs.K_diss_h = numerics.calc_diss_u(state, diss)

    """
    Meridional velocity
    """
    if settings.enable_hor_friction_cos_scaling:
        flux_east = update(
            flux_east,
            at[:-1],
            settings.A_h
            * vs.cosu[npx.newaxis, :, npx.newaxis] ** settings.hor_friction_cosPower
            * (vs.v[1:, :, :, vs.tau] - vs.v[:-1, :, :, vs.tau])
            / (vs.cosu * vs.dxu[:-1, npx.newaxis])[:, :, npx.newaxis]
            * vs.maskV[1:]
            * vs.maskV[:-1],
        )

        if settings.enable_noslip_lateral:
            flux_east = update_add(
                flux_east,
                at[:-1],
                2
                * settings.A_h
                * fxa[npx.newaxis, :, npx.newaxis]
                * vs.v[1:, :, :, vs.tau]
                / (vs.cosu * vs.dxu[:-1, npx.newaxis])[:, :, npx.newaxis]
                * vs.maskV[1:]
                * (1 - vs.maskV[:-1])
                - 2
                * settings.A_h
                * fxa[npx.newaxis, :, npx.newaxis]
                * vs.v[:-1, :, :, vs.tau]
                / (vs.cosu * vs.dxu[:-1, npx.newaxis])[:, :, npx.newaxis]
                * (1 - vs.maskV[1:])
                * vs.maskV[:-1],
            )

        flux_north = update(
            flux_north,
            at[:, :-1],
            settings.A_h
            * vs.cost[npx.newaxis, 1:, npx.newaxis] ** settings.hor_friction_cosPower
            * (vs.v[:, 1:, :, vs.tau] - vs.v[:, :-1, :, vs.tau])
            / vs.dyt[npx.newaxis, 1:, npx.newaxis]
            * vs.cost[npx.newaxis, 1:, npx.newaxis]
            * vs.maskV[:, :-1]
            * vs.maskV[:, 1:],
        )
    else:
        flux_east = update(
            flux_east,
            at[:-1],
            settings.A_h
            * (vs.v[1:, :, :, vs.tau] - vs.v[:-1, :, :, vs.tau])
            / (vs.cosu * vs.dxu[:-1, npx.newaxis])[:, :, npx.newaxis]
            * vs.maskV[1:]
            * vs.maskV[:-1],
        )

        if settings.enable_noslip_lateral:
            flux_east = update_add(
                flux_east,
                at[:-1],
                2
                * settings.A_h
                * vs.v[1:, :, :, vs.tau]
                / (vs.cosu * vs.dxu[:-1, npx.newaxis])[:, :, npx.newaxis]
                * vs.maskV[1:]
                * (1 - vs.maskV[:-1])
                - 2
                * settings.A_h
                * vs.v[:-1, :, :, vs.tau]
                / (vs.cosu * vs.dxu[:-1, npx.newaxis])[:, :, npx.newaxis]
                * (1 - vs.maskV[1:])
                * vs.maskV[:-1],
            )

        flux_north = update(
            flux_north,
            at[:, :-1],
            settings.A_h
            * (vs.v[:, 1:, :, vs.tau] - vs.v[:, :-1, :, vs.tau])
            / vs.dyt[npx.newaxis, 1:, npx.newaxis]
            * vs.cost[npx.newaxis, 1:, npx.newaxis]
            * vs.maskV[:, :-1]
            * vs.maskV[:, 1:],
        )

    flux_east = update(flux_east, at[-1, :, :], 0.0)
    flux_north = update(flux_north, at[:, -1, :], 0.0)

    """
    update tendency
    """
    vs.dv_mix = update_add(
        vs.dv_mix,
        at[2:-2, 2:-2],
        vs.maskV[2:-2, 2:-2]
        * (
            (flux_east[2:-2, 2:-2] - flux_east[1:-3, 2:-2])
            / (vs.cosu[2:-2] * vs.dxt[2:-2, npx.newaxis])[:, :, npx.newaxis]
            + (flux_north[2:-2, 2:-2] - flux_north[2:-2, 1:-3])
            / (vs.dyu[2:-2] * vs.cosu[2:-2])[npx.newaxis, :, npx.newaxis]
        ),
    )

    if settings.enable_conserve_energy:
        """
        diagnose dissipation by lateral friction
        """
        diss = update(
            diss,
            at[2:-2, 1:-2],
            0.5
            * (
                (vs.v[3:-1, 1:-2, :, vs.tau] - vs.v[2:-2, 1:-2, :, vs.tau]) * flux_east[2:-2, 1:-2]
                + (vs.v[2:-2, 1:-2, :, vs.tau] - vs.v[1:-3, 1:-2, :, vs.tau]) * flux_east[1:-3, 1:-2]
            )
            / (vs.cosu[1:-2] * vs.dxt[2:-2, npx.newaxis])[:, :, npx.newaxis]
            + 0.5
            * (
                (vs.v[2:-2, 2:-1, :, vs.tau] - vs.v[2:-2, 1:-2, :, vs.tau]) * flux_north[2:-2, 1:-2]
                + (vs.v[2:-2, 1:-2, :, vs.tau] - vs.v[2:-2, :-3, :, vs.tau]) * flux_north[2:-2, :-3]
            )
            / (vs.cosu[1:-2] * vs.dyu[1:-2])[npx.newaxis, :, npx.newaxis],
        )
        vs.K_diss_h = update_add(vs.K_diss_h, at[...], numerics.calc_diss_v(state, diss))

    return KernelOutput(du_mix=vs.du_mix, dv_mix=vs.dv_mix, K_diss_h=vs.K_diss_h)


@veros_kernel
def biharmonic_friction(state):
    """
    horizontal biharmonic friction
    dissipation is calculated and added to K_diss_h
    """
    vs = state.variables
    settings = state.settings

    flux_east = allocate(state.dimensions, ("xu", "yt", "zt"))
    flux_north = allocate(state.dimensions, ("xt", "yu", "zt"))
    visc = npx.sqrt(abs(settings.A_hbi))

    # each of these enters twice, so we halve the power
    cost_scaled = vs.cost ** (0.5 * settings.biharmonic_friction_cosPower)
    cosu_scaled = vs.cosu ** (0.5 * settings.biharmonic_friction_cosPower)

    """
    Zonal velocity
    """
    flux_east = update(
        flux_east,
        at[:-1, :, :],
        visc
        * cost_scaled[npx.newaxis, :, npx.newaxis]
        * (vs.u[1:, :, :, vs.tau] - vs.u[:-1, :, :, vs.tau])
        / (vs.cost[npx.newaxis, :, npx.newaxis] * vs.dxt[1:, npx.newaxis, npx.newaxis])
        * vs.maskU[1:, :, :]
        * vs.maskU[:-1, :, :],
    )
    flux_north = update(
        flux_north,
        at[:, :-1, :],
        visc
        * cosu_scaled[npx.newaxis, :-1, npx.newaxis]
        * (vs.u[:, 1:, :, vs.tau] - vs.u[:, :-1, :, vs.tau])
        / vs.dyu[npx.newaxis, :-1, npx.newaxis]
        * vs.maskU[:, 1:, :]
        * vs.maskU[:, :-1, :]
        * vs.cosu[npx.newaxis, :-1, npx.newaxis],
    )

    if settings.enable_noslip_lateral:
        flux_north = update_add(
            flux_north,
            at[:, :-1],
            2
            * visc
            * cosu_scaled[npx.newaxis, :-1, npx.newaxis]
            * vs.u[:, 1:, :, vs.tau]
            / vs.dyu[npx.newaxis, :-1, npx.newaxis]
            * vs.maskU[:, 1:]
            * (1 - vs.maskU[:, :-1])
            * vs.cosu[npx.newaxis, :-1, npx.newaxis]
            - 2
            * visc
            * cosu_scaled[npx.newaxis, :-1, npx.newaxis]
            * vs.u[:, :-1, :, vs.tau]
            / vs.dyu[npx.newaxis, :-1, npx.newaxis]
            * (1 - vs.maskU[:, 1:])
            * vs.maskU[:, :-1]
            * vs.cosu[npx.newaxis, :-1, npx.newaxis],
        )

    flux_east = update(flux_east, at[-1, :, :], 0.0)
    flux_north = update(flux_north, at[:, -1, :], 0.0)

    del2 = allocate(state.dimensions, ("xt", "yu", "zt"))
    del2 = update(
        del2,
        at[1:, 1:, :],
        (flux_east[1:, 1:, :] - flux_east[:-1, 1:, :])
        / (vs.cost[npx.newaxis, 1:, npx.newaxis] * vs.dxu[1:, npx.newaxis, npx.newaxis])
        + (flux_north[1:, 1:, :] - flux_north[1:, :-1, :])
        / (vs.cost[npx.newaxis, 1:, npx.newaxis] * vs.dyt[npx.newaxis, 1:, npx.newaxis]),
    )

    flux_east = update(
        flux_east,
        at[:-1, :, :],
        visc
        * cost_scaled[npx.newaxis, :, npx.newaxis]
        * (del2[1:, :, :] - del2[:-1, :, :])
        / (vs.cost[npx.newaxis, :, npx.newaxis] * vs.dxt[1:, npx.newaxis, npx.newaxis])
        * vs.maskU[1:, :, :]
        * vs.maskU[:-1, :, :],
    )
    flux_north = update(
        flux_north,
        at[:, :-1, :],
        visc
        * cosu_scaled[npx.newaxis, :-1, npx.newaxis]
        * (del2[:, 1:, :] - del2[:, :-1, :])
        / vs.dyu[npx.newaxis, :-1, npx.newaxis]
        * vs.maskU[:, 1:, :]
        * vs.maskU[:, :-1, :]
        * vs.cosu[npx.newaxis, :-1, npx.newaxis],
    )

    if settings.enable_noslip_lateral:
        flux_north = update_add(
            flux_north,
            at[:, :-1, :],
            2
            * visc
            * cosu_scaled[npx.newaxis, :-1, npx.newaxis]
            * del2[:, 1:, :]
            / vs.dyu[npx.newaxis, :-1, npx.newaxis]
            * vs.maskU[:, 1:, :]
            * (1 - vs.maskU[:, :-1, :])
            * vs.cosu[npx.newaxis, :-1, npx.newaxis]
            - 2
            * visc
            * cosu_scaled[npx.newaxis, :-1, npx.newaxis]
            * del2[:, :-1, :]
            / vs.dyu[npx.newaxis, :-1, npx.newaxis]
            * (1 - vs.maskU[:, 1:, :])
            * vs.maskU[:, :-1, :]
            * vs.cosu[npx.newaxis, :-1, npx.newaxis],
        )

    flux_east = update(flux_east, at[-1, :, :], 0.0)
    flux_north = update(flux_north, at[:, -1, :], 0.0)

    """
    update tendency
    """
    vs.du_mix = update_add(
        vs.du_mix,
        at[2:-2, 2:-2, :],
        -1
        * vs.maskU[2:-2, 2:-2, :]
        * (
            (flux_east[2:-2, 2:-2, :] - flux_east[1:-3, 2:-2, :])
            / (vs.cost[npx.newaxis, 2:-2, npx.newaxis] * vs.dxu[2:-2, npx.newaxis, npx.newaxis])
            + (flux_north[2:-2, 2:-2, :] - flux_north[2:-2, 1:-3, :])
            / (vs.cost[npx.newaxis, 2:-2, npx.newaxis] * vs.dyt[npx.newaxis, 2:-2, npx.newaxis])
        ),
    )

    if settings.enable_conserve_energy:
        """
        diagnose dissipation by lateral friction
        """
        flux_east = utilities.enforce_boundaries(flux_east, settings.enable_cyclic_x)
        flux_north = utilities.enforce_boundaries(flux_north, settings.enable_cyclic_x)
        diss = allocate(state.dimensions, ("xt", "yu", "zt"))
        diss = update(
            diss,
            at[1:-2, 2:-2, :],
            -0.5
            * (
                (vs.u[2:-1, 2:-2, :, vs.tau] - vs.u[1:-2, 2:-2, :, vs.tau]) * flux_east[1:-2, 2:-2, :]
                + (vs.u[1:-2, 2:-2, :, vs.tau] - vs.u[:-3, 2:-2, :, vs.tau]) * flux_east[:-3, 2:-2, :]
            )
            / (vs.cost[npx.newaxis, 2:-2, npx.newaxis] * vs.dxu[1:-2, npx.newaxis, npx.newaxis])
            - 0.5
            * (
                (vs.u[1:-2, 3:-1, :, vs.tau] - vs.u[1:-2, 2:-2, :, vs.tau]) * flux_north[1:-2, 2:-2, :]
                + (vs.u[1:-2, 2:-2, :, vs.tau] - vs.u[1:-2, 1:-3, :, vs.tau]) * flux_north[1:-2, 1:-3, :]
            )
            / (vs.cost[npx.newaxis, 2:-2, npx.newaxis] * vs.dyt[npx.newaxis, 2:-2, npx.newaxis]),
        )
        vs.K_diss_h = numerics.calc_diss_u(state, diss)

    """
    Meridional velocity
    """
    flux_east = update(
        flux_east,
        at[:-1, :, :],
        visc
        * cosu_scaled[npx.newaxis, :, npx.newaxis]
        * (vs.v[1:, :, :, vs.tau] - vs.v[:-1, :, :, vs.tau])
        / (vs.cosu[npx.newaxis, :, npx.newaxis] * vs.dxu[:-1, npx.newaxis, npx.newaxis])
        * vs.maskV[1:, :, :]
        * vs.maskV[:-1, :, :],
    )

    if settings.enable_noslip_lateral:
        flux_east = update_add(
            flux_east,
            at[:-1, :, :],
            2
            * visc
            * cosu_scaled[npx.newaxis, :, npx.newaxis]
            * vs.v[1:, :, :, vs.tau]
            / (vs.cosu[npx.newaxis, :, npx.newaxis] * vs.dxu[:-1, npx.newaxis, npx.newaxis])
            * vs.maskV[1:, :, :]
            * (1 - vs.maskV[:-1, :, :])
            - 2
            * visc
            * cosu_scaled[npx.newaxis, :, npx.newaxis]
            * vs.v[:-1, :, :, vs.tau]
            / (vs.cosu[npx.newaxis, :, npx.newaxis] * vs.dxu[:-1, npx.newaxis, npx.newaxis])
            * (1 - vs.maskV[1:, :, :])
            * vs.maskV[:-1, :, :],
        )

    flux_north = update(
        flux_north,
        at[:, :-1, :],
        visc
        * cost_scaled[npx.newaxis, :-1, npx.newaxis]
        * (vs.v[:, 1:, :, vs.tau] - vs.v[:, :-1, :, vs.tau])
        / vs.dyt[npx.newaxis, 1:, npx.newaxis]
        * vs.cost[npx.newaxis, 1:, npx.newaxis]
        * vs.maskV[:, :-1, :]
        * vs.maskV[:, 1:, :],
    )
    flux_east = update(flux_east, at[-1, :, :], 0.0)
    flux_north = update(flux_north, at[:, -1, :], 0.0)

    del2 = update(
        del2,
        at[1:, 1:, :],
        (flux_east[1:, 1:, :] - flux_east[:-1, 1:, :])
        / (vs.cosu[npx.newaxis, 1:, npx.newaxis] * vs.dxt[1:, npx.newaxis, npx.newaxis])
        + (flux_north[1:, 1:, :] - flux_north[1:, :-1, :])
        / (vs.dyu[npx.newaxis, 1:, npx.newaxis] * vs.cosu[npx.newaxis, 1:, npx.newaxis]),
    )

    flux_east = update(
        flux_east,
        at[:-1, :, :],
        visc
        * cosu_scaled[npx.newaxis, :, npx.newaxis]
        * (del2[1:, :, :] - del2[:-1, :, :])
        / (vs.cosu[npx.newaxis, :, npx.newaxis] * vs.dxu[:-1, npx.newaxis, npx.newaxis])
        * vs.maskV[1:, :, :]
        * vs.maskV[:-1, :, :],
    )

    if settings.enable_noslip_lateral:
        flux_east = update_add(
            flux_east,
            at[:-1, :, :],
            2
            * visc
            * cosu_scaled[npx.newaxis, :, npx.newaxis]
            * del2[1:, :, :]
            / (vs.cosu[npx.newaxis, :, npx.newaxis] * vs.dxu[:-1, npx.newaxis, npx.newaxis])
            * vs.maskV[1:, :, :]
            * (1 - vs.maskV[:-1, :, :])
            - 2
            * visc
            * cosu_scaled[npx.newaxis, :, npx.newaxis]
            * del2[:-1, :, :]
            / (vs.cosu[npx.newaxis, :, npx.newaxis] * vs.dxu[:-1, npx.newaxis, npx.newaxis])
            * (1 - vs.maskV[1:, :, :])
            * vs.maskV[:-1, :, :],
        )

    flux_north = update(
        flux_north,
        at[:, :-1, :],
        visc
        * cost_scaled[npx.newaxis, :-1, npx.newaxis]
        * (del2[:, 1:, :] - del2[:, :-1, :])
        / vs.dyt[npx.newaxis, 1:, npx.newaxis]
        * vs.cost[npx.newaxis, 1:, npx.newaxis]
        * vs.maskV[:, :-1, :]
        * vs.maskV[:, 1:, :],
    )
    flux_east = update(flux_east, at[-1, :, :], 0.0)
    flux_north = update(flux_north, at[:, -1, :], 0.0)

    """
    update tendency
    """
    vs.dv_mix = update_add(
        vs.dv_mix,
        at[2:-2, 2:-2, :],
        -1
        * vs.maskV[2:-2, 2:-2, :]
        * (
            (flux_east[2:-2, 2:-2, :] - flux_east[1:-3, 2:-2, :])
            / (vs.cosu[npx.newaxis, 2:-2, npx.newaxis] * vs.dxt[2:-2, npx.newaxis, npx.newaxis])
            + (flux_north[2:-2, 2:-2, :] - flux_north[2:-2, 1:-3, :])
            / (vs.dyu[npx.newaxis, 2:-2, npx.newaxis] * vs.cosu[npx.newaxis, 2:-2, npx.newaxis])
        ),
    )

    if settings.enable_conserve_energy:
        """
        diagnose dissipation by lateral friction
        """
        flux_east = utilities.enforce_boundaries(flux_east, settings.enable_cyclic_x)
        flux_north = utilities.enforce_boundaries(flux_north, settings.enable_cyclic_x)
        diss = update(
            diss,
            at[2:-2, 1:-2, :],
            -0.5
            * (
                (vs.v[3:-1, 1:-2, :, vs.tau] - vs.v[2:-2, 1:-2, :, vs.tau]) * flux_east[2:-2, 1:-2, :]
                + (vs.v[2:-2, 1:-2, :, vs.tau] - vs.v[1:-3, 1:-2, :, vs.tau]) * flux_east[1:-3, 1:-2, :]
            )
            / (vs.cosu[npx.newaxis, 1:-2, npx.newaxis] * vs.dxt[2:-2, npx.newaxis, npx.newaxis])
            - 0.5
            * (
                (vs.v[2:-2, 2:-1, :, vs.tau] - vs.v[2:-2, 1:-2, :, vs.tau]) * flux_north[2:-2, 1:-2, :]
                + (vs.v[2:-2, 1:-2, :, vs.tau] - vs.v[2:-2, :-3, :, vs.tau]) * flux_north[2:-2, :-3, :]
            )
            / (vs.cosu[npx.newaxis, 1:-2, npx.newaxis] * vs.dyu[npx.newaxis, 1:-2, npx.newaxis]),
        )
        vs.K_diss_h = update_add(vs.K_diss_h, at[...], numerics.calc_diss_v(state, diss))

    return KernelOutput(du_mix=vs.du_mix, dv_mix=vs.dv_mix, K_diss_h=vs.K_diss_h)


@veros_kernel
def momentum_sources(state):
    """
    other momentum sources
    dissipation is calculated and added to K_diss_bot
    """
    vs = state.variables
    settings = state.settings

    vs.du_mix = update_add(vs.du_mix, at[...], vs.maskU * vs.u_source)
    if settings.enable_conserve_energy:
        diss = -1 * vs.maskU * vs.u[..., vs.tau] * vs.u_source
        vs.K_diss_bot = update_add(vs.K_diss_bot, at[...], numerics.calc_diss_u(state, diss))

    vs.dv_mix = update_add(vs.dv_mix, at[...], vs.maskV * vs.v_source)
    if settings.enable_conserve_energy:
        diss = -1 * vs.maskV * vs.v[..., vs.tau] * vs.v_source
        vs.K_diss_bot = update_add(vs.K_diss_bot, at[...], numerics.calc_diss_v(state, diss))

    return KernelOutput(du_mix=vs.du_mix, dv_mix=vs.dv_mix, K_diss_bot=vs.K_diss_bot)


@veros_routine
def friction(state):
    vs = state.variables
    settings = state.settings

    """
    vertical friction
    """
    vs.K_diss_v = update(vs.K_diss_v, at[...], 0.0)

    if settings.enable_implicit_vert_friction:
        vs.update(implicit_vert_friction(state))

    if settings.enable_explicit_vert_friction:
        vs.update(explicit_vert_friction(state))

    """
    TEM formalism for eddy-driven velocity
    """
    if settings.enable_TEM_friction:
        vs.update(isoneutral.isoneutral_friction(state))

    """
    horizontal friction
    """
    if settings.enable_hor_friction:
        vs.update(harmonic_friction(state))

    if settings.enable_biharmonic_friction:
        vs.update(biharmonic_friction(state))

    """
    Rayleigh and bottom friction
    """
    vs.K_diss_bot = update(vs.K_diss_bot, at[...], 0.0)

    if settings.enable_ray_friction:
        vs.update(rayleigh_friction(state))

    if settings.enable_bottom_friction:
        vs.update(linear_bottom_friction(state))

    if settings.enable_quadratic_bottom_friction:
        vs.update(quadratic_bottom_friction(state))

    """
    add user defined forcing
    """
    if settings.enable_momentum_sources:
        vs.update(momentum_sources(state))



from veros import logger, veros_kernel, veros_routine, KernelOutput
from veros.variables import allocate
from veros.distributed import global_max
from veros.core import utilities as mainutils
from veros.core.operators import numpy as npx, update, at
from veros.core.external import island, line_integrals, solve_stream
from veros.core.external.solvers import get_linear_solver


@veros_routine
def get_isleperim(state):
    """
    preprocess land map using MOMs algorithm for B-grid to determine number of islands
    """
    from veros.state import resize_dimension

    vs = state.variables

    island.isleperim(state)

    # now that we know the number of islands we can resize
    # all arrays depending on that
    nisle = int(global_max(npx.max(vs.land_map)))
    resize_dimension(state, "isle", nisle)
    vs.isle = npx.arange(nisle)


@veros_routine
def streamfunction_init(state):
    """
    prepare for island integrals
    """
    vs = state.variables
    settings = state.settings

    logger.info("Initializing streamfunction method")

    get_isleperim(state)

    vs.update(boundary_masks(state))

    # populate linear solver cache
    linear_solver = get_linear_solver(state)

    """
    precalculate time independent boundary components of streamfunction
    """
    forc = allocate(state.dimensions, ("xt", "yt"))

    vs.psin = update(vs.psin, at[...], vs.maskZ[..., -1, npx.newaxis])

    for isle in range(state.dimensions["isle"]):
        logger.info(f" Solving for boundary contribution by island {isle:d}")
        isle_boundary = (
            vs.line_dir_east_mask[..., isle]
            | vs.line_dir_west_mask[..., isle]
            | vs.line_dir_north_mask[..., isle]
            | vs.line_dir_south_mask[..., isle]
        )
        isle_sol = linear_solver.solve(state, forc, vs.psin[:, :, isle], boundary_val=isle_boundary)
        vs.psin = update(vs.psin, at[:, :, isle], isle_sol)

    vs.psin = mainutils.enforce_boundaries(vs.psin, settings.enable_cyclic_x)

    line_psin_out = island_integrals(state)
    vs.update(line_psin_out)

    """
    take care of initial velocity
    """

    # transfer initial velocity to tendency
    vs.du = update(vs.du, at[..., vs.tau], vs.u[..., vs.tau] / settings.dt_mom / (1.5 + settings.AB_eps))
    vs.dv = update(vs.dv, at[..., vs.tau], vs.v[..., vs.tau] / settings.dt_mom / (1.5 + settings.AB_eps))
    vs.u = update(vs.u, at[...], 0)
    vs.v = update(vs.v, at[...], 0)

    # run streamfunction solver to determine initial barotropic and baroclinic modes
    solve_stream.solve_streamfunction(state)

    vs.psi = update(vs.psi, at[...], vs.psi[..., vs.taup1, npx.newaxis])
    vs.u = update(
        vs.u, at[...], mainutils.enforce_boundaries(vs.u[..., vs.taup1, npx.newaxis], settings.enable_cyclic_x)
    )
    vs.v = update(
        vs.v, at[...], mainutils.enforce_boundaries(vs.v[..., vs.taup1, npx.newaxis], settings.enable_cyclic_x)
    )
    vs.du = update(vs.du, at[..., vs.tau], 0)
    vs.dv = update(vs.dv, at[..., vs.tau], 0)


@veros_kernel
def island_integrals(state):
    """
    precalculate time independent island integrals
    """
    vs = state.variables

    uloc = allocate(state.dimensions, ("xt", "yt", "isle"))
    vloc = allocate(state.dimensions, ("xt", "yt", "isle"))

    uloc = update(
        uloc,
        at[1:, 1:, :],
        -(vs.psin[1:, 1:, :] - vs.psin[1:, :-1, :])
        * vs.maskU[1:, 1:, -1, npx.newaxis]
        / vs.dyt[npx.newaxis, 1:, npx.newaxis]
        * vs.hur[1:, 1:, npx.newaxis],
    )

    vloc = update(
        vloc,
        at[1:, 1:, ...],
        (vs.psin[1:, 1:, :] - vs.psin[:-1, 1:, :])
        * vs.maskV[1:, 1:, -1, npx.newaxis]
        / (vs.cosu[npx.newaxis, 1:, npx.newaxis] * vs.dxt[1:, npx.newaxis, npx.newaxis])
        * vs.hvr[1:, 1:, npx.newaxis],
    )

    vs.line_psin = line_integrals.line_integrals(state, uloc=uloc, vloc=vloc, kind="full")
    return KernelOutput(line_psin=vs.line_psin)


@veros_kernel
def boundary_masks(state):
    """
    now that the number of islands is known we can allocate the rest of the variables
    """
    vs = state.variables
    settings = state.settings

    boundary_map = vs.land_map[..., npx.newaxis] == npx.arange(1, state.dimensions["isle"] + 1)

    if settings.enable_cyclic_x:
        vs.line_dir_east_mask = update(
            vs.line_dir_east_mask, at[2:-2, 1:-1], boundary_map[3:-1, 1:-1] & ~boundary_map[3:-1, 2:]
        )
        vs.line_dir_west_mask = update(
            vs.line_dir_west_mask, at[2:-2, 1:-1], boundary_map[2:-2, 2:] & ~boundary_map[2:-2, 1:-1]
        )
        vs.line_dir_south_mask = update(
            vs.line_dir_south_mask, at[2:-2, 1:-1], boundary_map[2:-2, 1:-1] & ~boundary_map[3:-1, 1:-1]
        )
        vs.line_dir_north_mask = update(
            vs.line_dir_north_mask, at[2:-2, 1:-1], boundary_map[3:-1, 2:] & ~boundary_map[2:-2, 2:]
        )
    else:
        vs.line_dir_east_mask = update(
            vs.line_dir_east_mask, at[1:-1, 1:-1], boundary_map[2:, 1:-1] & ~boundary_map[2:, 2:]
        )
        vs.line_dir_west_mask = update(
            vs.line_dir_west_mask, at[1:-1, 1:-1], boundary_map[1:-1, 2:] & ~boundary_map[1:-1, 1:-1]
        )
        vs.line_dir_south_mask = update(
            vs.line_dir_south_mask, at[1:-1, 1:-1], boundary_map[1:-1, 1:-1] & ~boundary_map[2:, 1:-1]
        )
        vs.line_dir_north_mask = update(
            vs.line_dir_north_mask, at[1:-1, 1:-1], boundary_map[2:, 2:] & ~boundary_map[1:-1, 2:]
        )

    vs.isle_boundary_mask = ~npx.any(
        vs.line_dir_east_mask | vs.line_dir_west_mask | vs.line_dir_south_mask | vs.line_dir_north_mask, axis=2
    )

    return KernelOutput(
        isle_boundary_mask=vs.isle_boundary_mask,
        line_dir_east_mask=vs.line_dir_east_mask,
        line_dir_west_mask=vs.line_dir_west_mask,
        line_dir_south_mask=vs.line_dir_south_mask,
        line_dir_north_mask=vs.line_dir_north_mask,
    )



import sys
import warnings


LOGLEVELS = ("trace", "debug", "info", "warning", "error")


def _inject_proc_rank(record):
    from veros import runtime_state

    return record["extra"].update(proc_rank=runtime_state.proc_rank)


def setup_logging(loglevel="info", stream_sink=sys.stdout, log_all_processes=False):
    from loguru import logger

    handler_conf = dict(
        sink=stream_sink,
        level=loglevel.upper(),
        colorize=sys.stdout.isatty(),
    )

    if not hasattr(logger, "diagnostic"):
        logger.level("DIAGNOSTIC", no=45)

    logger.level("TRACE", color="<dim>")
    logger.level("DEBUG", color="<dim><cyan>")
    logger.level("INFO", color="")
    logger.level("SUCCESS", color="<dim><green>")
    logger.level("WARNING", color="<yellow>")
    logger.level("ERROR", color="<bold><red>")
    logger.level("DIAGNOSTIC", color="<bold><yellow>")
    logger.level("CRITICAL", color="<bold><red><WHITE>")

    logger = logger.patch(_inject_proc_rank)
    if log_all_processes:
        handler_conf.update(format="{extra[proc_rank]} | <level>{message}</level>")
    else:
        handler_conf.update(format="<level>{message}</level>", filter=lambda record: record["extra"]["proc_rank"] == 0)

    def diagnostic(_, message, *args, **kwargs):
        logger.opt(depth=1).log("DIAGNOSTIC", message, *args, **kwargs)

    logger.__class__.diagnostic = diagnostic

    def showwarning(message, cls, source, lineno, *args):
        logger.warning(
            "{warning}: {message} ({source}:{lineno})",
            message=message,
            warning=cls.__name__,
            source=source,
            lineno=lineno,
        )

    warnings.showwarning = showwarning

    logger.configure(handlers=[handler_conf])
    logger.enable("veros")

    return logger



import cython
from cpython.pycapsule cimport PyCapsule_New

from libc.stdint cimport int32_t, int64_t


@cython.cdivision(True)
cdef void _tdma_cython_double(int32_t n, double* a, double* b, double* c, double* d, double* cp, double* dp) nogil:
    cdef:
        int32_t i
        double denom

    if n < 1:
        return

    cp[0] = c[0] / b[0]
    dp[0] = d[0] / b[0]

    for i in range(1, n):
        denom = 1. / (b[i] - a[i] * cp[i - 1])
        cp[i] = c[i] * denom
        dp[i] = (d[i] - a[i] * dp[i - 1]) * denom

    for i in range(n - 2, -1, -1):
        dp[i] -= cp[i] * dp[i + 1]


@cython.cdivision(True)
cdef void _tdma_cython_float(int32_t n, float* a, float* b, float* c, float* d, float* cp, float* dp) nogil:
    cdef:
        int32_t i
        float denom

    if n < 1:
        return

    cp[0] = c[0] / b[0]
    dp[0] = d[0] / b[0]

    for i in range(1, n):
        denom = 1. / (b[i] - a[i] * cp[i - 1])
        cp[i] = c[i] * denom
        dp[i] = (d[i] - a[i] * dp[i - 1]) * denom

    for i in range(n - 2, -1, -1):
        dp[i] -= cp[i] * dp[i + 1]


cdef void tdma_cython_double(void** out_ptr, void** data_ptr) nogil:
    cdef:
        int64_t i, j, system_depth, system_start
        int64_t ii = 0

        # decode inputs
        double* a = (<double*>data_ptr[0])
        double* b = (<double*>data_ptr[1])
        double* c = (<double*>data_ptr[2])
        double* d = (<double*>data_ptr[3])
        int32_t* system_depths = (<int32_t*>data_ptr[4])
        int64_t num_systems = (<int64_t*>data_ptr[5])[0]
        int64_t stride = (<int64_t*>data_ptr[6])[0]

        double* out = (<double*>out_ptr[0])
        double* workspace = (<double*>out_ptr[1])

    for i in range(num_systems):
        system_depth = system_depths[i]
        system_start = stride - system_depth

        for j in range(system_start):
            out[ii + j] = 0.

        _tdma_cython_double(
            system_depth,
            &a[ii + system_start],
            &b[ii + system_start],
            &c[ii + system_start],
            &d[ii + system_start],
            workspace,
            &out[ii + system_start],
        )

        ii += stride


cdef void tdma_cython_float(void** out_ptr, void** data_ptr) nogil:
    cdef:
        int64_t i, j, system_depth, system_start
        int64_t ii = 0

        # decode inputs
        float* a = (<float*>data_ptr[0])
        float* b = (<float*>data_ptr[1])
        float* c = (<float*>data_ptr[2])
        float* d = (<float*>data_ptr[3])
        int32_t* system_depths = (<int32_t*>data_ptr[4])
        int64_t num_systems = (<int64_t*>data_ptr[5])[0]
        int64_t stride = (<int64_t*>data_ptr[6])[0]

        float* out = (<float*>out_ptr[0])
        float* workspace = (<float*>out_ptr[1])

    for i in range(num_systems):
        system_depth = system_depths[i]
        system_start = stride - system_depth

        for j in range(system_start):
            out[ii + j] = 0.0

        _tdma_cython_float(
            system_depth,
            &a[ii + system_start],
            &b[ii + system_start],
            &c[ii + system_start],
            &d[ii + system_start],
            workspace,
            &out[ii + system_start],
        )

        ii += stride


cpu_custom_call_targets = {}

cdef register_custom_call_target(unicode fn_name, void* fn):
    cdef const char* name = 'xla._CUSTOM_CALL_TARGET'
    cpu_custom_call_targets[fn_name] = PyCapsule_New(fn, name, NULL)

register_custom_call_target('tdma_cython_double', <void*>(tdma_cython_double))
register_custom_call_target('tdma_cython_float', <void*>(tdma_cython_float))



from veros.core.isoneutral.isoneutral import (  # noqa: F401
    check_isoneutral_slope_crit,
    isoneutral_diffusion_pre,
    isoneutral_diag_streamfunction,
)

from veros.core.isoneutral.diffusion import (  # noqa: F401
    isoneutral_diffusion,
    isoneutral_skew_diffusion,
)

from veros.core.isoneutral.friction import (  # noqa: F401
    isoneutral_friction,
)



import json
import datetime
import threading
import contextlib

import numpy as np

from veros import (
    logger,
    variables,
    distributed,
    runtime_state,
    runtime_settings as rs,
    __version__ as veros_version,
)

"""
netCDF output is designed to follow the COARDS guidelines from
http://ferret.pmel.noaa.gov/Ferret/documentation/coards-netcdf-conventions
"""


def _get_setup_code(pyfile):
    try:
        with open(pyfile, "r") as f:
            return f.read()
    except FileNotFoundError:
        return "UNKNOWN"


def initialize_file(state, ncfile, extra_dimensions=None, create_time_dimension=True):
    """
    Define standard grid in netcdf file
    """
    import h5netcdf

    if not isinstance(ncfile, h5netcdf.File):
        raise TypeError("Argument needs to be a netCDF4 Dataset")

    if rs.setup_file is None:
        setup_file = "UNKNOWN"
        setup_code = "UNKNOWN"
    else:
        setup_file = rs.setup_file
        setup_code = _get_setup_code(rs.setup_file)

    ncfile.attrs.update(
        date_created=datetime.datetime.today().isoformat(),
        veros_version=veros_version,
        setup_identifier=state.settings.identifier,
        setup_description=state.settings.description,
        setup_settings=json.dumps(state.settings.todict()),
        setup_file=setup_file,
        setup_code=setup_code,
    )

    dimensions = dict(state.dimensions)
    if extra_dimensions is not None:
        dimensions.update(extra_dimensions)

    for dim in dimensions:
        # time steps are peeled off explicitly
        if dim in variables.TIMESTEPS:
            continue

        if dim in state.var_meta:
            var = state.var_meta[dim]

            # skip inactive dimensions
            if not var.active:
                continue

            var_data = getattr(state.variables, dim)
        else:
            # create dummy variable for dimensions without data
            var = variables.Variable(dim, (dim,), time_dependent=False)
            var_data = np.arange(dimensions[dim])

        dimsize = variables.get_shape(dimensions, var.dims[::-1], include_ghosts=False, local=False)[0]
        ncfile.dimensions[dim] = dimsize
        initialize_variable(state, dim, var, ncfile)
        write_variable(state, dim, var, var_data, ncfile)

    if create_time_dimension:
        ncfile.dimensions["Time"] = None
        nc_dim_var_time = ncfile.create_variable("Time", ("Time",), float)
        nc_dim_var_time.attrs.update(
            long_name="Time",
            units="days",
            time_origin="01-JAN-1900 00:00:00",
        )


def initialize_variable(state, key, var, ncfile):
    if var.dims is None:
        dims = ()
    else:
        dims = tuple(d for d in var.dims if d in ncfile.dimensions)

    if var.time_dependent and "Time" in ncfile.dimensions:
        dims += ("Time",)

    if key in ncfile.variables:
        logger.warning(f"Variable {key} already initialized")
        return

    kwargs = {}
    if rs.hdf5_gzip_compression and runtime_state.proc_num == 1:
        kwargs.update(compression="gzip", compression_opts=1)

    chunksize = [
        variables.get_shape(state.dimensions, (d,), local=True, include_ghosts=False)[0] if d in state.dimensions else 1
        for d in dims
    ]

    dtype = var.dtype
    if dtype is None:
        dtype = rs.float_type
    elif dtype == "bool":
        dtype = "uint8"

    fillvalue = variables.get_fill_value(dtype)

    # transpose all dimensions in netCDF output (convention in most ocean models)
    v = ncfile.create_variable(key, dims[::-1], dtype, fillvalue=fillvalue, chunks=tuple(chunksize[::-1]), **kwargs)
    v.missing_value = fillvalue
    v.attrs.update(long_name=var.name, units=var.units, **var.extra_attributes)


def advance_time(time_value, ncfile):
    current_time_step = len(ncfile.variables["Time"])
    ncfile.resize_dimension("Time", current_time_step + 1)
    ncfile.variables["Time"][current_time_step] = time_value


def add_dimension(dim, dim_size, ncfile):
    ncfile.dimensions[dim] = int(dim_size)


def write_variable(state, key, var, var_data, ncfile, time_step=-1):
    var_data = var_data * var.scale

    gridmask = var.get_mask(state.settings, state.variables)
    if gridmask is not None:
        newaxes = (slice(None),) * gridmask.ndim + (np.newaxis,) * (var_data.ndim - gridmask.ndim)
        var_data = np.where(gridmask.astype("bool")[newaxes], var_data, variables.get_fill_value(var_data.dtype))

    if var.dims:
        tmask = tuple(state.variables.tau if dim in variables.TIMESTEPS else slice(None) for dim in var.dims)
        var_data = variables.remove_ghosts(var_data, var.dims)[tmask].T

    var_obj = ncfile.variables[key]

    nx, ny = state.dimensions["xt"], state.dimensions["yt"]
    chunk, _ = distributed.get_chunk_slices(nx, ny, var_obj.dimensions)

    if "Time" in var_obj.dimensions:
        assert var_obj.dimensions[0] == "Time"
        chunk = (time_step,) + chunk[1:]

    var_obj[chunk] = var_data


@contextlib.contextmanager
def threaded_io(filepath, mode):
    """
    If using IO threads, start a new thread to write the netCDF data to disk.
    """
    import h5py
    import h5netcdf

    if rs.use_io_threads:
        _wait_for_disk(filepath)
        _io_locks[filepath].clear()

    kwargs = dict()

    if int(h5py.__version__.split(".")[0]) >= 3:
        kwargs.update(decode_vlen_strings=True)

    if runtime_state.proc_num > 1:
        kwargs.update(driver="mpio", comm=rs.mpi_comm)

    nc_dataset = h5netcdf.File(filepath, mode, **kwargs)

    try:
        yield nc_dataset

    finally:
        if rs.use_io_threads:
            threading.Thread(target=_write_to_disk, args=(nc_dataset, filepath)).start()
        else:
            _write_to_disk(nc_dataset, filepath)


_io_locks = {}


def _add_to_locks(file_id):
    """
    If there is no lock for file_id, create one
    """
    if file_id not in _io_locks:
        _io_locks[file_id] = threading.Event()
        _io_locks[file_id].set()


def _wait_for_disk(file_id):
    """
    Wait for the lock of file_id to be released
    """
    logger.debug(f"Waiting for lock {file_id} to be released")
    _add_to_locks(file_id)
    lock_released = _io_locks[file_id].wait(rs.io_timeout)

    if not lock_released:
        raise RuntimeError("Timeout while waiting for disk IO to finish")


def _write_to_disk(ncfile, file_id):
    """
    Sync netCDF data to disk, close file handle, and release lock.
    May run in a separate thread.
    """
    try:
        ncfile.close()
    finally:
        if rs.use_io_threads and file_id is not None:
            _io_locks[file_id].set()



import os

from veros import veros_kernel, KernelOutput, runtime_settings
from veros.core.operators import numpy as npx, update_multiply, at
from veros.diagnostics.base import VerosDiagnostic
from veros.variables import Variable
from veros.distributed import global_sum


ENERGY_VARIABLES = dict(
    nitts=Variable("nitts", None, write_to_restart=True),
    # mean energy content
    k_m=Variable("Mean kinetic energy", None, "J", "Mean kinetic energy", write_to_restart=True),
    Hd_m=Variable("Mean dynamic enthalpy", None, "J", "Mean dynamic enthalpy", write_to_restart=True),
    eke_m=Variable("Meso-scale eddy energy", None, "J", "Meso-scale eddy energy", write_to_restart=True),
    iw_m=Variable("Internal wave energy", None, "J", "Internal wave energy", write_to_restart=True),
    tke_m=Variable("Turbulent kinetic energy", None, "J", "Turbulent kinetic energy", write_to_restart=True),
    # energy changes
    dE_tot_m=Variable("Change of total energy", None, "W", "Change of total energy", write_to_restart=True),
    dk_m=Variable("Change of KE", None, "W", "Change of kinetic energy", write_to_restart=True),
    dHd_m=Variable("Change of Hd", None, "W", "Change of dynamic enthalpy", write_to_restart=True),
    deke_m=Variable("Change of EKE", None, "W", "Change of meso-scale eddy energy", write_to_restart=True),
    diw_m=Variable("Change of E_iw", None, "W", "Change of internal wave energy", write_to_restart=True),
    dtke_m=Variable("Change of TKE", None, "W", "Change of tubulent kinetic energy", write_to_restart=True),
    # dissipation
    ke_diss_m=Variable("Dissipation of KE", None, "W", "Dissipation of kinetic energy", write_to_restart=True),
    Hd_diss_m=Variable("Dissipation of Hd", None, "W", "Dissipation of dynamic enthalpy", write_to_restart=True),
    eke_diss_m=Variable(
        "Dissipation of EKE", None, "W", "Dissipation of meso-scale eddy energy", write_to_restart=True
    ),
    iw_diss_m=Variable("Dissipation of E_iw", None, "W", "Dissipation of internal wave energy", write_to_restart=True),
    tke_diss_m=Variable(
        "Dissipation of TKE", None, "W", "Dissipation of turbulent kinetic energy", write_to_restart=True
    ),
    adv_diss_m=Variable("Dissipation by advection", None, "W", "Dissipation by advection", write_to_restart=True),
    # external forcing
    wind_m=Variable("Wind work", None, "W", "Wind work", write_to_restart=True),
    dHd_sources_m=Variable(
        "Hd production by ext. sources",
        None,
        "W",
        "Dynamic enthalpy production through external sources",
        write_to_restart=True,
    ),
    iw_forc_m=Variable(
        "External forcing of E_iw", None, "W", "External forcing of internal wave energy", write_to_restart=True
    ),
    tke_forc_m=Variable(
        "External forcing of TKE", None, "W", "External forcing of turbulent kinetic energy", write_to_restart=True
    ),
    # exchange
    ke_hd_m=Variable(
        "Exchange KE -> Hd", None, "W", "Exchange between kinetic energy and dynamic enthalpy", write_to_restart=True
    ),
    ke_tke_m=Variable(
        "Exchange KE -> TKE by vert. friction",
        None,
        "W",
        "Exchange between kinetic energy and turbulent kinetic energy by vertical friction",
        write_to_restart=True,
    ),
    ke_iw_m=Variable(
        "Exchange KE -> IW by bottom friction",
        None,
        "W",
        "Exchange between kinetic energy and internal wave energy by bottom friction",
        write_to_restart=True,
    ),
    tke_hd_m=Variable(
        "Exchange TKE -> Hd by vertical mixing",
        None,
        "W",
        "Exchange between turbulent kinetic energy and dynamic enthalpy by vertical mixing",
        write_to_restart=True,
    ),
    ke_eke_m=Variable(
        "Exchange KE -> EKE by lateral friction",
        None,
        "W",
        "Exchange between kinetic energy and eddy kinetic energy by lateral friction",
        write_to_restart=True,
    ),
    hd_eke_m=Variable(
        "Exchange Hd -> EKE by GM and lateral mixing",
        None,
        "W",
        "Exchange between dynamic enthalpy and eddy kinetic energy by GM and lateral mixing",
        write_to_restart=True,
    ),
    eke_tke_m=Variable(
        "Exchange EKE -> TKE", None, "W", "Exchange between eddy and turbulent kinetic energy", write_to_restart=True
    ),
    eke_iw_m=Variable(
        "Exchange EKE -> IW",
        None,
        "W",
        "Exchange between eddy kinetic energy and internal wave energy",
        write_to_restart=True,
    ),
    # cabbeling
    cabb_m=Variable("Cabbeling by vertical mixing", None, "W", "Cabbeling by vertical mixing", write_to_restart=True),
    cabb_iso_m=Variable(
        "Cabbeling by isopycnal mixing", None, "W", "Cabbeling by isopycnal mixing", write_to_restart=True
    ),
)


DEFAULT_OUTPUT_VARS = [var for var in ENERGY_VARIABLES.keys() if var not in ("nitts",)]


class Energy(VerosDiagnostic):
    """Diagnose globally averaged energy cycle. Also averages energy in time."""

    name = "energy"  #:
    output_path = "{identifier}.energy.nc"  #: File to write to. May contain format strings that are replaced with Veros attributes.
    output_frequency = None  #: Frequency (in seconds) in which output is written.
    sampling_frequency = None  #: Frequency (in seconds) in which variables are accumulated.

    var_meta = ENERGY_VARIABLES

    def __init__(self, state):
        self.output_variables = DEFAULT_OUTPUT_VARS.copy()

    def initialize(self, state):
        self.initialize_variables(state)
        self.initialize_output(state)

    def diagnose(self, state):
        energies = diagnose_kernel(state)

        # store results
        for energy, val in energies._asdict().items():
            total_val = self.variables.get(energy)
            setattr(self.variables, energy, total_val + val)

        self.variables.nitts = self.variables.nitts + 1

    def output(self, state):
        if not os.path.isfile(self.get_output_file_name(state)):
            self.initialize_output(state)

        energy_vs = self.variables
        nitts = float(energy_vs.nitts or 1)

        for key in self.output_variables:
            val = getattr(energy_vs, key)
            setattr(energy_vs, key, val * state.settings.rho_0 / nitts)

        self.write_output(state)

        for key in self.output_variables:
            setattr(energy_vs, key, 0.0)

        energy_vs.nitts = 0


@veros_kernel
def diagnose_kernel(state):
    vs = state.variables
    settings = state.settings

    # changes of dynamic enthalpy
    vol_t = vs.area_t[2:-2, 2:-2, npx.newaxis] * vs.dzt[npx.newaxis, npx.newaxis, :] * vs.maskT[2:-2, 2:-2, :]

    dP_iso = global_sum(
        npx.sum(
            vol_t
            * settings.grav
            / settings.rho_0
            * (
                -vs.int_drhodT[2:-2, 2:-2, :, vs.tau] * vs.dtemp_iso[2:-2, 2:-2, :]
                - vs.int_drhodS[2:-2, 2:-2, :, vs.tau] * vs.dsalt_iso[2:-2, 2:-2, :]
            )
        )
    )

    dP_hmix = global_sum(
        npx.sum(
            vol_t
            * settings.grav
            / settings.rho_0
            * (
                -vs.int_drhodT[2:-2, 2:-2, :, vs.tau] * vs.dtemp_hmix[2:-2, 2:-2, :]
                - vs.int_drhodS[2:-2, 2:-2, :, vs.tau] * vs.dsalt_hmix[2:-2, 2:-2, :]
            )
        )
    )

    dP_vmix = global_sum(
        npx.sum(
            vol_t
            * settings.grav
            / settings.rho_0
            * (
                -vs.int_drhodT[2:-2, 2:-2, :, vs.tau] * vs.dtemp_vmix[2:-2, 2:-2, :]
                - vs.int_drhodS[2:-2, 2:-2, :, vs.tau] * vs.dsalt_vmix[2:-2, 2:-2, :]
            )
        )
    )

    dP_m = global_sum(
        npx.sum(
            vol_t
            * settings.grav
            / settings.rho_0
            * (
                -vs.int_drhodT[2:-2, 2:-2, :, vs.tau] * vs.dtemp[2:-2, 2:-2, :, vs.tau]
                - vs.int_drhodS[2:-2, 2:-2, :, vs.tau] * vs.dsalt[2:-2, 2:-2, :, vs.tau]
            )
        )
    )

    dP_m_all = dP_m + dP_vmix + dP_hmix + dP_iso

    # changes of kinetic energy
    vol_u = vs.area_u[2:-2, 2:-2, npx.newaxis] * vs.dzt[npx.newaxis, npx.newaxis, :]
    vol_v = vs.area_v[2:-2, 2:-2, npx.newaxis] * vs.dzt[npx.newaxis, npx.newaxis, :]
    k_m = global_sum(
        npx.sum(
            vol_t
            * 0.5
            * (
                0.5 * (vs.u[2:-2, 2:-2, :, vs.tau] ** 2 + vs.u[1:-3, 2:-2, :, vs.tau] ** 2)
                + 0.5 * (vs.v[2:-2, 2:-2, :, vs.tau] ** 2)
                + vs.v[2:-2, 1:-3, :, vs.tau] ** 2
            )
        )
    )
    p_m = global_sum(npx.sum(vol_t * vs.Hd[2:-2, 2:-2, :, vs.tau]))
    dk_m = global_sum(
        npx.sum(
            vs.u[2:-2, 2:-2, :, vs.tau] * vs.du[2:-2, 2:-2, :, vs.tau] * vol_u
            + vs.v[2:-2, 2:-2, :, vs.tau] * vs.dv[2:-2, 2:-2, :, vs.tau] * vol_v
            + vs.u[2:-2, 2:-2, :, vs.tau] * vs.du_mix[2:-2, 2:-2, :] * vol_u
            + vs.v[2:-2, 2:-2, :, vs.tau] * vs.dv_mix[2:-2, 2:-2, :] * vol_v
        )
    )

    # K*Nsqr and KE and dyn. enthalpy dissipation
    vol_w = vs.area_t[2:-2, 2:-2, npx.newaxis] * vs.dzw[npx.newaxis, npx.newaxis, :] * vs.maskW[2:-2, 2:-2, :]
    vol_w = update_multiply(vol_w, at[:, :, -1], 0.5)

    def mean_w(var):
        return global_sum(npx.sum(var[2:-2, 2:-2, :] * vol_w))

    mdiss_vmix = mean_w(vs.P_diss_v)
    mdiss_nonlin = mean_w(vs.P_diss_nonlin)
    mdiss_adv = mean_w(vs.P_diss_adv)
    mdiss_hmix = mean_w(vs.P_diss_hmix)
    mdiss_iso = mean_w(vs.P_diss_iso)
    mdiss_skew = mean_w(vs.P_diss_skew)
    mdiss_sources = mean_w(vs.P_diss_sources)

    mdiss_h = mean_w(vs.K_diss_h)
    mdiss_v = mean_w(vs.K_diss_v)
    mdiss_gm = mean_w(vs.K_diss_gm)
    mdiss_bot = mean_w(vs.K_diss_bot)

    wrhom = global_sum(
        npx.sum(
            -vs.area_t[2:-2, 2:-2, npx.newaxis]
            * vs.maskW[2:-2, 2:-2, :-1]
            * (vs.p_hydro[2:-2, 2:-2, 1:] - vs.p_hydro[2:-2, 2:-2, :-1])
            * vs.w[2:-2, 2:-2, :-1, vs.tau]
        )
    )

    # wind work
    if runtime_settings.pyom_compatibility_mode:
        # surface_tau* has different units in PyOM
        wind = global_sum(
            npx.sum(
                vs.u[2:-2, 2:-2, -1, vs.tau]
                * vs.surface_taux[2:-2, 2:-2]
                * vs.maskU[2:-2, 2:-2, -1]
                * vs.area_u[2:-2, 2:-2]
                + vs.v[2:-2, 2:-2, -1, vs.tau]
                * vs.surface_tauy[2:-2, 2:-2]
                * vs.maskV[2:-2, 2:-2, -1]
                * vs.area_v[2:-2, 2:-2]
            )
        )
    else:
        wind = global_sum(
            npx.sum(
                vs.u[2:-2, 2:-2, -1, vs.tau]
                * vs.surface_taux[2:-2, 2:-2]
                / settings.rho_0
                * vs.maskU[2:-2, 2:-2, -1]
                * vs.area_u[2:-2, 2:-2]
                + vs.v[2:-2, 2:-2, -1, vs.tau]
                * vs.surface_tauy[2:-2, 2:-2]
                / settings.rho_0
                * vs.maskV[2:-2, 2:-2, -1]
                * vs.area_v[2:-2, 2:-2]
            )
        )

    # meso-scale energy
    if settings.enable_eke:
        eke_m = mean_w(vs.eke[..., vs.tau])
        deke_m = global_sum(
            npx.sum(vol_w * (vs.eke[2:-2, 2:-2, :, vs.taup1] - vs.eke[2:-2, 2:-2, :, vs.tau]) / settings.dt_tracer)
        )
        eke_diss = mean_w(vs.eke_diss_iw)
        eke_diss_tke = mean_w(vs.eke_diss_tke)
    else:
        eke_m = deke_m = eke_diss_tke = 0.0
        eke_diss = mdiss_gm + mdiss_h + mdiss_skew
        if not settings.enable_store_cabbeling_heat:
            eke_diss += -mdiss_hmix - mdiss_iso

    # small-scale energy
    if settings.enable_tke:
        dt_tke = settings.dt_mom
        tke_m = mean_w(vs.tke[..., vs.tau])
        dtke_m = mean_w((vs.tke[..., vs.taup1] - vs.tke[..., vs.tau]) / dt_tke)
        tke_diss = mean_w(vs.tke_diss)
        tke_forc = global_sum(
            npx.sum(
                vs.area_t[2:-2, 2:-2]
                * vs.maskW[2:-2, 2:-2, -1]
                * (vs.forc_tke_surface[2:-2, 2:-2] + vs.tke_surf_corr[2:-2, 2:-2])
            )
        )
    else:
        tke_m = dtke_m = tke_diss = tke_forc = 0.0

    # internal wave energy
    if settings.enable_idemix:
        iw_m = mean_w(vs.E_iw[..., vs.tau])
        diw_m = global_sum(
            npx.sum(vol_w * (vs.E_iw[2:-2, 2:-2, :, vs.taup1] - vs.E_iw[2:-2, 2:-2, :, vs.tau]) / vs.dt_tracer)
        )
        iw_diss = mean_w(vs.iw_diss)

        k = npx.maximum(1, vs.kbot[2:-2, 2:-2]) - 1
        mask = k[:, :, npx.newaxis] == npx.arange(settings.nz)[npx.newaxis, npx.newaxis, :]
        iwforc = global_sum(
            npx.sum(
                vs.area_t[2:-2, 2:-2]
                * (
                    vs.forc_iw_surface[2:-2, 2:-2] * vs.maskW[2:-2, 2:-2, -1]
                    + npx.sum(mask * vs.forc_iw_bottom[2:-2, 2:-2, npx.newaxis] * vs.maskW[2:-2, 2:-2, :], axis=2)
                )
            )
        )
    else:
        iw_m = diw_m = iwforc = 0.0
        iw_diss = eke_diss

    if settings.enable_store_bottom_friction_tke:
        ke_tke_m = mdiss_v + mdiss_bot
        ke_iw_m = 0.0
    else:
        ke_tke_m = mdiss_v
        ke_iw_m = mdiss_bot

    hd_eke_m = -mdiss_skew
    tke_hd_m = -mdiss_vmix - mdiss_adv

    if not settings.enable_store_cabbeling_heat:
        hd_eke_m = hd_eke_m - mdiss_hmix - mdiss_iso
        tke_hd_m = tke_hd_m - mdiss_nonlin

    return KernelOutput(
        k_m=k_m,
        Hd_m=p_m,
        eke_m=eke_m,
        iw_m=iw_m,
        tke_m=tke_m,
        dk_m=dk_m,
        dHd_m=dP_m_all + mdiss_sources,
        deke_m=deke_m,
        diw_m=diw_m,
        dtke_m=dtke_m,
        dE_tot_m=dk_m + dP_m_all + mdiss_sources + deke_m + diw_m + dtke_m,
        wind_m=wind,
        dHd_sources_m=mdiss_sources,
        iw_forc_m=iwforc,
        tke_forc_m=tke_forc,
        ke_diss_m=mdiss_h + mdiss_v + mdiss_gm + mdiss_bot,
        Hd_diss_m=mdiss_vmix + mdiss_nonlin + mdiss_hmix + mdiss_adv + mdiss_iso + mdiss_skew,
        eke_diss_m=eke_diss + eke_diss_tke,
        iw_diss_m=iw_diss,
        tke_diss_m=tke_diss,
        adv_diss_m=mdiss_adv,
        ke_hd_m=wrhom,
        ke_eke_m=mdiss_h + mdiss_gm,
        hd_eke_m=-mdiss_skew,
        ke_tke_m=ke_tke_m,
        ke_iw_m=ke_iw_m,
        tke_hd_m=tke_hd_m,
        eke_tke_m=eke_diss_tke,
        eke_iw_m=eke_diss,
        cabb_m=mdiss_nonlin,
        cabb_iso_m=mdiss_hmix + mdiss_iso,
    )



#!/usr/bin/env python

import functools
import subprocess
import shlex
import sys
import os
import time

import click

LAST_N_FILENAME = "{identifier}.current_run"
CHILD_TIMEOUT = 10
POLL_DELAY = 0.1


class ShellCommand(click.ParamType):
    name = "command"

    def convert(self, value, param, ctx):
        return shlex.split(value)


def get_current_n(filename):
    if not os.path.isfile(filename):
        return 0

    with open(filename, "r") as f:
        return int(f.read())


def write_next_n(n, filename):
    with open(filename, "w") as f:
        f.write(str(n))


def unparse(args):
    return shlex.join(args)


def call_veros(cmd, name, n, runlen):
    identifier = f"{name}.{n:0>4}"
    prev_id = f"{name}.{n - 1:0>4}"
    args = [
        "-s",
        "identifier",
        identifier,
        "-s",
        "restart_output_filename",
        "{identifier}.restart.h5",
        "-s",
        "runlen",
        str(runlen),
    ]
    if n:
        args += ["-s", "restart_input_filename", f"{prev_id}.restart.h5"]

    # make sure this isn't buffered
    sys.stdout.write(f'\n >>> {" ".join(cmd + args)}\n\n')
    sys.stdout.flush()

    try:
        subprocess.check_call(unparse(cmd + args), shell=True)
    except subprocess.CalledProcessError:
        raise RuntimeError(f"Run {n} failed, exiting")


def resubmit(identifier, num_runs, length_per_run, veros_cmd, callback):
    """Performs several runs of Veros back to back, using the previous run as restart input.

    Intended to be used with scheduling systems (e.g. SLURM or PBS).

    """
    last_n_filename = LAST_N_FILENAME.format(identifier=identifier)

    current_n = get_current_n(last_n_filename)
    if current_n >= num_runs:
        return

    call_veros(veros_cmd, identifier, current_n, length_per_run)

    next_n = current_n + 1
    write_next_n(next_n, last_n_filename)

    if next_n >= num_runs:
        return

    next_proc = subprocess.Popen(unparse(callback), shell=True)

    # catch immediately crashing processes
    timeout = CHILD_TIMEOUT

    while timeout > 0:
        retcode = next_proc.poll()
        if retcode is not None:
            if retcode > 0:
                # process crashed
                raise RuntimeError(f"Callback exited with {retcode}")
            else:
                break
        time.sleep(POLL_DELAY)
        timeout -= POLL_DELAY


@click.command("veros-resubmit", short_help="Re-run a Veros setup several times")
@click.option("-i", "--identifier", required=True, help="Base identifier of the simulation")
@click.option("-n", "--num-runs", type=click.INT, required=True, help="Total number of runs to execute")
@click.option("-l", "--length-per-run", type=click.FLOAT, required=True, help="Length (in seconds) of each run")
@click.option(
    "-c", "--veros-cmd", type=ShellCommand(), required=True, help="The command that is used to call veros (quoted)"
)
@click.option(
    "--callback",
    metavar="CMD",
    type=ShellCommand(),
    default=None,
    help="Command to call after each run has finished (quoted, default: call self)",
)
@functools.wraps(resubmit)
def cli(*args, **kwargs):
    if kwargs["callback"] is None:
        kwargs["callback"] = sys.argv

    resubmit(*args, **kwargs)



from veros import veros_kernel, veros_routine, KernelOutput
from veros.variables import allocate
from veros.distributed import global_and
from veros.core import density, diffusion, utilities
from veros.core.operators import update, at, numpy as npx


@veros_kernel
def u_centered_grid(dyt, dyu, yt, yu):
    yu = update(yu, at[0], 0)
    yu = update(yu, at[1:], npx.cumsum(dyt[1:]))

    yt = update(yt, at[0], yu[0] - dyt[0] * 0.5)
    yt = update(yt, at[1:], 2 * yu[:-1])

    alternating_pattern = npx.ones_like(yt)
    alternating_pattern = update(alternating_pattern, at[::2], -1)
    yt = update(yt, at[...], alternating_pattern * npx.cumsum(alternating_pattern * yt))

    dyu = update(dyu, at[:-1], yt[1:] - yt[:-1])
    dyu = update(dyu, at[-1], 2 * dyt[-1] - dyu[-2])
    return dyu, yt, yu


@veros_kernel
def calc_grid_spacings_kernel(state):
    vs = state.variables
    settings = state.settings

    if settings.enable_cyclic_x:
        vs.dxt = update(vs.dxt, at[-2:], vs.dxt[2:4])
        vs.dxt = update(vs.dxt, at[:2], vs.dxt[-4:-2])
    else:
        vs.dxt = update(vs.dxt, at[-2:], vs.dxt[-3])
        vs.dxt = update(vs.dxt, at[:2], vs.dxt[2])

    vs.dyt = update(vs.dyt, at[-2:], vs.dyt[-3])
    vs.dyt = update(vs.dyt, at[:2], vs.dyt[2])

    """
    grid in east/west direction
    """
    vs.dxu, vs.xt, vs.xu = u_centered_grid(vs.dxt, vs.dxu, vs.xt, vs.xu)
    vs.xt = vs.xt + settings.x_origin - vs.xu[2]
    vs.xu = vs.xu + settings.x_origin - vs.xu[2]

    if settings.enable_cyclic_x:
        vs.xt = update(vs.xt, at[-2:], vs.xt[2:4])
        vs.xt = update(vs.xt, at[:2], vs.xt[-4:-2])
        vs.xu = update(vs.xu, at[-2:], vs.xt[2:4])
        vs.xu = update(vs.xu, at[:2], vs.xu[-4:-2])
        vs.dxu = update(vs.dxu, at[-2:], vs.dxu[2:4])
        vs.dxu = update(vs.dxu, at[:2], vs.dxu[-4:-2])

    """
    grid in north/south direction
    """
    vs.dyu, vs.yt, vs.yu = u_centered_grid(vs.dyt, vs.dyu, vs.yt, vs.yu)
    vs.yt = vs.yt + settings.y_origin - vs.yu[2]
    vs.yu = vs.yu + settings.y_origin - vs.yu[2]

    if settings.coord_degree:
        """
        convert from degrees to pseudo cartesian grid
        """
        vs.dxt = vs.dxt * settings.degtom
        vs.dxu = vs.dxu * settings.degtom
        vs.dyt = vs.dyt * settings.degtom
        vs.dyu = vs.dyu * settings.degtom

    """
    grid in vertical direction
    """
    vs.dzw, vs.zt, vs.zw = u_centered_grid(vs.dzt, vs.dzw, vs.zt, vs.zw)
    vs.zt = vs.zt - vs.zw[-1]
    vs.zw = vs.zw - vs.zw[-1]  # enforce 0 boundary height

    return KernelOutput(
        dxt=vs.dxt,
        dyt=vs.dyt,
        dxu=vs.dxu,
        dyu=vs.dyu,
        xt=vs.xt,
        yt=vs.yt,
        xu=vs.xu,
        yu=vs.yu,
        dzw=vs.dzw,
        zt=vs.zt,
        zw=vs.zw,
    )


@veros_routine(
    # all inputs are 1D, so doing this on the main process should be fine
    dist_safe=False,
    local_variables=(
        "dxt",
        "dxu",
        "xt",
        "xu",
        "dyt",
        "dyu",
        "yt",
        "yu",
        "dzt",
        "dzw",
        "zt",
        "zw",
    ),
)
def calc_grid_spacings(state):
    vs = state.variables
    vs.update(calc_grid_spacings_kernel(state))


@veros_kernel
def calc_grid_metrics_kernel(state):
    vs = state.variables
    settings = state.settings

    """
    metric factors
    """
    if settings.coord_degree:
        vs.cost = update(vs.cost, at[...], npx.cos(vs.yt * settings.pi / 180.0))
        vs.cosu = update(vs.cosu, at[...], npx.cos(vs.yu * settings.pi / 180.0))
        vs.tantr = update(vs.tantr, at[...], npx.tan(vs.yt * settings.pi / 180.0) / settings.radius)
    else:
        vs.cost = update(vs.cost, at[...], 1.0)
        vs.cosu = update(vs.cosu, at[...], 1.0)
        vs.tantr = update(vs.tantr, at[...], 0.0)

    """
    precalculate area of boxes
    """
    vs.area_t = update(vs.area_t, at[...], vs.cost * vs.dyt * vs.dxt[:, npx.newaxis])
    vs.area_u = update(vs.area_u, at[...], vs.cost * vs.dyt * vs.dxu[:, npx.newaxis])
    vs.area_v = update(vs.area_v, at[...], vs.cosu * vs.dyu * vs.dxt[:, npx.newaxis])

    return KernelOutput(
        cost=vs.cost,
        cosu=vs.cosu,
        tantr=vs.tantr,
        area_t=vs.area_t,
        area_u=vs.area_u,
        area_v=vs.area_v,
    )


@veros_routine
def calc_grid(state):
    """
    setup grid based on dxt,dyt,dzt and x_origin, y_origin
    """
    calc_grid_spacings(state)

    vs = state.variables
    vs.update(calc_grid_metrics_kernel(state))


@veros_routine
def calc_beta(state):
    """
    calculate beta = df/dy
    """
    vs = state.variables
    settings = state.settings
    vs.beta = update(
        vs.beta,
        at[:, 2:-2],
        0.5
        * (
            (vs.coriolis_t[:, 3:-1] - vs.coriolis_t[:, 2:-2]) / vs.dyu[2:-2]
            + (vs.coriolis_t[:, 2:-2] - vs.coriolis_t[:, 1:-3]) / vs.dyu[1:-3]
        ),
    )
    vs.beta = utilities.enforce_boundaries(vs.beta, settings.enable_cyclic_x)


@veros_kernel
def calc_topo_kernel(state):
    vs = state.variables
    settings = state.settings

    """
    close domain
    """
    vs.kbot = update(vs.kbot, at[:, :2], 0)
    vs.kbot = update(vs.kbot, at[:, -2:], 0)

    vs.kbot = utilities.enforce_boundaries(vs.kbot, settings.enable_cyclic_x)

    if not settings.enable_cyclic_x:
        vs.kbot = update(vs.kbot, at[:2, :], 0)
        vs.kbot = update(vs.kbot, at[-2:, :], 0)

    """
    Land masks
    """
    land_mask = vs.kbot > 0
    ks = npx.arange(vs.maskT.shape[2])[npx.newaxis, npx.newaxis, :]

    vs.maskT = update(vs.maskT, at[...], land_mask[..., npx.newaxis] & (vs.kbot[..., npx.newaxis] - 1 <= ks))
    vs.maskT = utilities.enforce_boundaries(vs.maskT, settings.enable_cyclic_x)

    vs.maskU = update(vs.maskU, at[...], vs.maskT)
    vs.maskU = update(vs.maskU, at[:-1, :, :], npx.minimum(vs.maskT[:-1, :, :], vs.maskT[1:, :, :]))
    vs.maskU = utilities.enforce_boundaries(vs.maskU, settings.enable_cyclic_x)

    vs.maskV = update(vs.maskV, at[...], vs.maskT)
    vs.maskV = update(vs.maskV, at[:, :-1], npx.minimum(vs.maskT[:, :-1], vs.maskT[:, 1:]))
    vs.maskV = utilities.enforce_boundaries(vs.maskV, settings.enable_cyclic_x)

    vs.maskZ = update(vs.maskZ, at[...], vs.maskT)
    vs.maskZ = update(
        vs.maskZ, at[:-1, :-1], npx.minimum(npx.minimum(vs.maskT[:-1, :-1], vs.maskT[:-1, 1:]), vs.maskT[1:, :-1])
    )
    vs.maskZ = utilities.enforce_boundaries(vs.maskZ, settings.enable_cyclic_x)

    vs.maskW = update(vs.maskW, at[...], vs.maskT)
    vs.maskW = update(vs.maskW, at[:, :, :-1], npx.minimum(vs.maskT[:, :, :-1], vs.maskT[:, :, 1:]))

    """
    total depth
    """
    vs.ht = npx.sum(vs.maskT * vs.dzt[npx.newaxis, npx.newaxis, :], axis=2)
    vs.hu = npx.sum(vs.maskU * vs.dzt[npx.newaxis, npx.newaxis, :], axis=2)
    vs.hv = npx.sum(vs.maskV * vs.dzt[npx.newaxis, npx.newaxis, :], axis=2)

    vs.hur = npx.where(vs.hu != 0, 1 / (vs.hu + 1e-22), 0)
    vs.hvr = npx.where(vs.hv != 0, 1 / (vs.hv + 1e-22), 0)

    return KernelOutput(
        maskT=vs.maskT,
        maskU=vs.maskU,
        maskV=vs.maskV,
        maskW=vs.maskW,
        maskZ=vs.maskZ,
        ht=vs.ht,
        hu=vs.hu,
        hv=vs.hv,
        hur=vs.hur,
        hvr=vs.hvr,
        kbot=vs.kbot,
    )


@veros_routine
def calc_topo(state):
    """
    calulate masks, total depth etc
    """
    vs = state.variables
    vs.update(calc_topo_kernel(state))


@veros_kernel
def calc_initial_conditions_kernel(state):
    vs = state.variables
    settings = state.settings

    vs.temp = utilities.enforce_boundaries(vs.temp, settings.enable_cyclic_x)
    vs.salt = utilities.enforce_boundaries(vs.salt, settings.enable_cyclic_x)

    vs.rho = density.get_rho(state, vs.salt, vs.temp, npx.abs(vs.zt)[:, npx.newaxis]) * vs.maskT[..., npx.newaxis]
    vs.Hd = (
        density.get_dyn_enthalpy(state, vs.salt, vs.temp, npx.abs(vs.zt)[:, npx.newaxis]) * vs.maskT[..., npx.newaxis]
    )
    vs.int_drhodT = update(
        vs.int_drhodT, at[...], density.get_int_drhodT(state, vs.salt, vs.temp, npx.abs(vs.zt)[:, npx.newaxis])
    )
    vs.int_drhodS = update(
        vs.int_drhodS, at[...], density.get_int_drhodS(state, vs.salt, vs.temp, npx.abs(vs.zt)[:, npx.newaxis])
    )

    fxa = -settings.grav / settings.rho_0 / vs.dzw[npx.newaxis, npx.newaxis, :] * vs.maskW
    vs.Nsqr = update(
        vs.Nsqr,
        at[:, :, :-1, :],
        fxa[:, :, :-1, npx.newaxis]
        * (
            density.get_rho(state, vs.salt[:, :, 1:, :], vs.temp[:, :, 1:, :], npx.abs(vs.zt)[:-1, npx.newaxis])
            - vs.rho[:, :, :-1, :]
        ),
    )
    vs.Nsqr = update(vs.Nsqr, at[:, :, -1, :], vs.Nsqr[:, :, -2, :])

    return KernelOutput(
        salt=vs.salt,
        temp=vs.temp,
        rho=vs.rho,
        Hd=vs.Hd,
        int_drhodT=vs.int_drhodT,
        int_drhodS=vs.int_drhodS,
        Nsqr=vs.Nsqr,
    )


@veros_routine
def calc_initial_conditions(state):
    """
    calculate dyn. enthalp, etc
    """
    vs = state.variables

    if npx.any(vs.salt < 0.0):
        raise RuntimeError("encountered negative salinity")

    vs.update(calc_initial_conditions_kernel(state))


@veros_kernel
def ugrid_to_tgrid(state, a):
    vs = state.variables
    b = update(
        a,
        at[2:-2, :, :],
        (
            vs.dxu[2:-2, npx.newaxis, npx.newaxis] * a[2:-2, :, :]
            + vs.dxu[1:-3, npx.newaxis, npx.newaxis] * a[1:-3, :, :]
        )
        / (2 * vs.dxt[2:-2, npx.newaxis, npx.newaxis]),
    )
    return b


@veros_kernel
def vgrid_to_tgrid(state, a):
    vs = state.variables
    b = update(
        a,
        at[:, 2:-2, :],
        (vs.area_v[:, 2:-2, npx.newaxis] * a[:, 2:-2, :] + vs.area_v[:, 1:-3, npx.newaxis] * a[:, 1:-3, :])
        / (2 * vs.area_t[:, 2:-2, npx.newaxis]),
    )
    return b


@veros_kernel
def calc_diss_u(state, diss):
    vs = state.variables
    ks = allocate(state.dimensions, ("xt", "yt"))
    ks = update(ks, at[1:-2, 2:-2], npx.maximum(vs.kbot[1:-2, 2:-2], vs.kbot[2:-1, 2:-2]))
    diss_u = diffusion.dissipation_on_wgrid(state, diss, ks)
    return ugrid_to_tgrid(state, diss_u)


@veros_kernel
def calc_diss_v(state, diss):
    vs = state.variables
    ks = allocate(state.dimensions, ("xt", "yt"))
    ks = update(ks, at[2:-2, 1:-2], npx.maximum(vs.kbot[2:-2, 1:-2], vs.kbot[2:-2, 2:-1]))
    diss_v = diffusion.dissipation_on_wgrid(state, diss, ks)
    return vgrid_to_tgrid(state, diss_v)


@veros_kernel
def sanity_check(state):
    return global_and(npx.all(npx.isfinite(state.variables.u)))



# cython: language=c++

from cpython.pycapsule cimport PyCapsule_New


cdef extern from "cuda_runtime_api.h":
    ctypedef void* cudaStream_t
    cdef struct TridiagDescriptor:
        int num_systems
        int system_depth


cdef extern from "cuda_tdma_kernels.h":
    void CudaTridiagFloat(cudaStream_t stream, void** buffers, const char* opaque, size_t opaque_len)
    void CudaTridiagDouble(cudaStream_t stream, void** buffers, const char* opaque, size_t opaque_len)


cpdef bytes build_tridiag_descriptor(int num_systems, int system_depth):
    cdef TridiagDescriptor desc = TridiagDescriptor(num_systems, system_depth)
    return bytes((<char*> &desc)[:sizeof(TridiagDescriptor)])


gpu_custom_call_targets = {}

cdef register_custom_call_target(unicode fn_name, void* fn):
    cdef const char* name = "xla._CUSTOM_CALL_TARGET"
    gpu_custom_call_targets[fn_name] = PyCapsule_New(fn, name, NULL)

register_custom_call_target("tdma_cuda_float", <void*>(CudaTridiagFloat))
register_custom_call_target("tdma_cuda_double", <void*>(CudaTridiagDouble))



from veros.core.operators import numpy as npx

from veros import veros_kernel, veros_routine, KernelOutput
from veros.variables import allocate
from veros.core import utilities, diffusion
from veros.core.operators import update, update_add, at


@veros_kernel
def _calc_tracer_fluxes(state, tr, K_iso, K_skew):
    vs = state.variables

    tr_pad = utilities.pad_z_edges(tr[..., vs.tau])

    K1 = K_iso - K_skew
    K2 = K_iso + K_skew

    flux_east = allocate(state.dimensions, ("xt", "yt", "zt"))
    flux_north = allocate(state.dimensions, ("xt", "yt", "zt"))
    flux_top = allocate(state.dimensions, ("xt", "yt", "zt"))

    """
    construct total isoneutral tracer flux at east face of 'T' cells
    """
    diffloc = allocate(state.dimensions, ("xt", "yt", "zt"))[1:-2, 2:-2]
    diffloc = update(
        diffloc,
        at[:, :, 1:],
        0.25 * (K1[1:-2, 2:-2, 1:] + K1[1:-2, 2:-2, :-1] + K1[2:-1, 2:-2, 1:] + K1[2:-1, 2:-2, :-1]),
    )
    diffloc = update(diffloc, at[:, :, 0], 0.5 * (K1[1:-2, 2:-2, 0] + K1[2:-1, 2:-2, 0]))

    sumz = 0.0
    for kr in range(2):
        for ip in range(2):
            sumz = sumz + diffloc * vs.Ai_ez[1:-2, 2:-2, :, ip, kr] * (
                tr_pad[1 + ip : -2 + ip, 2:-2, 1 + kr : -1 + kr or None] - tr_pad[1 + ip : -2 + ip, 2:-2, kr : -2 + kr]
            )

    flux_east = update(
        flux_east,
        at[1:-2, 2:-2, :],
        sumz / (4.0 * vs.dzt[npx.newaxis, npx.newaxis, :])
        + (tr[2:-1, 2:-2, :, vs.tau] - tr[1:-2, 2:-2, :, vs.tau])
        / (vs.cost[npx.newaxis, 2:-2, npx.newaxis] * vs.dxu[1:-2, npx.newaxis, npx.newaxis])
        * vs.K_11[1:-2, 2:-2, :],
    )

    """
    construct total isoneutral tracer flux at north face of 'T' cells
    """
    diffloc = allocate(state.dimensions, ("xt", "yt", "zt"))[2:-2, 1:-2]
    diffloc = update(
        diffloc,
        at[:, :, 1:],
        0.25 * (K1[2:-2, 1:-2, 1:] + K1[2:-2, 1:-2, :-1] + K1[2:-2, 2:-1, 1:] + K1[2:-2, 2:-1, :-1]),
    )
    diffloc = update(diffloc, at[:, :, 0], 0.5 * (K1[2:-2, 1:-2, 0] + K1[2:-2, 2:-1, 0]))

    sumz = 0.0
    for kr in range(2):
        for jp in range(2):
            sumz = sumz + diffloc * vs.Ai_nz[2:-2, 1:-2, :, jp, kr] * (
                tr_pad[2:-2, 1 + jp : -2 + jp, 1 + kr : -1 + kr or None] - tr_pad[2:-2, 1 + jp : -2 + jp, kr : -2 + kr]
            )

    flux_north = update(
        flux_north,
        at[2:-2, 1:-2, :],
        vs.cosu[npx.newaxis, 1:-2, npx.newaxis]
        * (
            sumz / (4.0 * vs.dzt[npx.newaxis, npx.newaxis, :])
            + (tr[2:-2, 2:-1, :, vs.tau] - tr[2:-2, 1:-2, :, vs.tau])
            / vs.dyu[npx.newaxis, 1:-2, npx.newaxis]
            * vs.K_22[2:-2, 1:-2, :]
        ),
    )

    """
    compute the vertical tracer flux 'flux_top' containing the K31
    and K32 components which are to be solved explicitly. The K33
    component will be treated implicitly. Note that there are some
    cancellations of dxu(i-1+ip) and dyu(jrow-1+jp)
    """
    diffloc = K2[2:-2, 2:-2, :-1]
    sumx = 0.0

    for ip in range(2):
        for kr in range(2):
            sumx = sumx + diffloc * vs.Ai_bx[2:-2, 2:-2, :-1, ip, kr] / vs.cost[npx.newaxis, 2:-2, npx.newaxis] * (
                tr[2 + ip : -2 + ip, 2:-2, kr : -1 + kr or None, vs.tau]
                - tr[1 + ip : -3 + ip, 2:-2, kr : -1 + kr or None, vs.tau]
            )

    sumy = 0.0
    for jp in range(2):
        for kr in range(2):
            sumy = sumy + diffloc * vs.Ai_by[2:-2, 2:-2, :-1, jp, kr] * vs.cosu[
                npx.newaxis, 1 + jp : -3 + jp, npx.newaxis
            ] * (
                tr[2:-2, 2 + jp : -2 + jp, kr : -1 + kr or None, vs.tau]
                - tr[2:-2, 1 + jp : -3 + jp, kr : -1 + kr or None, vs.tau]
            )

    flux_top = update(
        flux_top,
        at[2:-2, 2:-2, :-1],
        sumx / (4 * vs.dxt[2:-2, npx.newaxis, npx.newaxis])
        + sumy / (4 * vs.dyt[npx.newaxis, 2:-2, npx.newaxis] * vs.cost[npx.newaxis, 2:-2, npx.newaxis]),
    )
    flux_top = update(flux_top, at[:, :, -1], 0.0)

    return flux_east, flux_north, flux_top


@veros_kernel
def _calc_explicit_part(state, flux_east, flux_north, flux_top):
    vs = state.variables

    explicit_part = allocate(state.dimensions, ("xt", "yt", "zt"))
    explicit_part = update(
        explicit_part,
        at[2:-2, 2:-2, :],
        vs.maskT[2:-2, 2:-2, :]
        * (
            (flux_east[2:-2, 2:-2, :] - flux_east[1:-3, 2:-2, :])
            / (vs.cost[npx.newaxis, 2:-2, npx.newaxis] * vs.dxt[2:-2, npx.newaxis, npx.newaxis])
            + (flux_north[2:-2, 2:-2, :] - flux_north[2:-2, 1:-3, :])
            / (vs.cost[npx.newaxis, 2:-2, npx.newaxis] * vs.dyt[npx.newaxis, 2:-2, npx.newaxis])
        ),
    )
    explicit_part = update_add(explicit_part, at[:, :, 0], vs.maskT[:, :, 0] * flux_top[:, :, 0] / vs.dzt[0])
    explicit_part = update_add(
        explicit_part,
        at[:, :, 1:],
        vs.maskT[:, :, 1:] * (flux_top[:, :, 1:] - flux_top[:, :, :-1]) / vs.dzt[npx.newaxis, npx.newaxis, 1:],
    )

    return explicit_part


@veros_kernel
def _calc_implicit_part(state, tr):
    vs = state.variables
    settings = state.settings

    _, water_mask, edge_mask = utilities.create_water_masks(vs.kbot[2:-2, 2:-2], settings.nz)

    a_tri = allocate(state.dimensions, ("xt", "yt", "zt"))[2:-2, 2:-2]
    b_tri = allocate(state.dimensions, ("xt", "yt", "zt"))[2:-2, 2:-2]
    c_tri = allocate(state.dimensions, ("xt", "yt", "zt"))[2:-2, 2:-2]
    delta = allocate(state.dimensions, ("xt", "yt", "zt"))[2:-2, 2:-2]

    delta = update(
        delta, at[:, :, :-1], settings.dt_tracer / vs.dzw[npx.newaxis, npx.newaxis, :-1] * vs.K_33[2:-2, 2:-2, :-1]
    )
    delta = update(delta, at[:, :, -1], 0.0)
    a_tri = update(a_tri, at[:, :, 1:], -delta[:, :, :-1] / vs.dzt[npx.newaxis, npx.newaxis, 1:])
    b_tri = update(
        b_tri, at[:, :, 1:-1], 1 + (delta[:, :, 1:-1] + delta[:, :, :-2]) / vs.dzt[npx.newaxis, npx.newaxis, 1:-1]
    )
    b_tri = update(b_tri, at[:, :, -1], 1 + delta[:, :, -2] / vs.dzt[npx.newaxis, npx.newaxis, -1])
    b_tri_edge = 1 + (delta[:, :, :] / vs.dzt[npx.newaxis, npx.newaxis, :])
    c_tri = update(c_tri, at[:, :, :-1], -delta[:, :, :-1] / vs.dzt[npx.newaxis, npx.newaxis, :-1])
    sol = utilities.solve_implicit(
        a_tri, b_tri, c_tri, tr[2:-2, 2:-2, :, vs.taup1], water_mask, b_edge=b_tri_edge, edge_mask=edge_mask
    )
    implicit_part = npx.where(water_mask, sol, tr[2:-2, 2:-2, :, vs.taup1])
    return implicit_part


@veros_kernel(static_args=("iso", "skew"))
def isoneutral_diffusion_tracer(state, tr, dtracer_iso, iso=True, skew=False):
    """
    Isoneutral diffusion for general tracers
    """
    vs = state.variables
    settings = state.settings

    if iso:
        K_iso = vs.K_iso
    else:
        K_iso = 0.0

    if skew:
        K_skew = vs.K_gm
    else:
        K_skew = 0.0

    flux_east, flux_north, flux_top = _calc_tracer_fluxes(state, tr, K_iso, K_skew)

    """
    add explicit part
    """
    dtr = _calc_explicit_part(state, flux_east, flux_north, flux_top)
    dtracer_iso = dtracer_iso + dtr
    tr = update_add(tr, at[2:-2, 2:-2, :, vs.taup1], settings.dt_tracer * dtr[2:-2, 2:-2, :])

    """
    add implicit part
    """
    if iso:
        new_tr = update(tr, at[2:-2, 2:-2, :, vs.taup1], _calc_implicit_part(state, tr))
        dtracer_iso = dtracer_iso + (new_tr[:, :, :, vs.taup1] - tr[:, :, :, vs.taup1]) / settings.dt_tracer
        tr = new_tr

    return tr, dtracer_iso, flux_east, flux_north, flux_top


@veros_kernel(static_args=("istemp", "iso"))
def isoneutral_diffusion_kernel(state, tr, istemp, iso=True):
    vs = state.variables
    settings = state.settings

    if istemp:
        dtracer_iso = vs.dtemp_iso
    else:
        dtracer_iso = vs.dsalt_iso

    tr, dtracer_iso, flux_east, flux_north, flux_top = isoneutral_diffusion_tracer(
        state, tr, dtracer_iso, iso=iso, skew=not iso
    )

    out = {}

    if istemp:
        out.update(temp=tr, dtemp_iso=dtracer_iso)
    else:
        out.update(salt=tr, dsalt_iso=dtracer_iso)

    """
    dissipation by isopycnal mixing
    """
    if settings.enable_conserve_energy:
        if istemp:
            int_drhodX = vs.int_drhodT[:, :, :, vs.tau]
        else:
            int_drhodX = vs.int_drhodS[:, :, :, vs.tau]

        """
        dissipation interpolated on W-grid
        """
        diss = diffusion.compute_dissipation(state, int_drhodX, flux_east, flux_north)
        diss_wgrid = diffusion.dissipation_on_wgrid(state, diss, vs.kbot)

        if not iso:
            vs.P_diss_skew = vs.P_diss_skew + diss_wgrid
        else:
            vs.P_diss_iso = vs.P_diss_iso + diss_wgrid

        """
        diagnose dissipation of dynamic enthalpy by explicit and implicit vertical mixing
        """
        fxa = (-int_drhodX[2:-2, 2:-2, 1:] + int_drhodX[2:-2, 2:-2, :-1]) / vs.dzw[npx.newaxis, npx.newaxis, :-1]

        if not iso:
            vs.P_diss_skew = update_add(
                vs.P_diss_skew,
                at[2:-2, 2:-2, :-1],
                -settings.grav / settings.rho_0 * fxa * flux_top[2:-2, 2:-2, :-1] * vs.maskW[2:-2, 2:-2, :-1],
            )

            out["P_diss_skew"] = vs.P_diss_skew

        else:
            vs.P_diss_iso = update_add(
                vs.P_diss_iso,
                at[2:-2, 2:-2, :-1],
                -settings.grav
                / settings.rho_0
                * fxa
                * (
                    flux_top[2:-2, 2:-2, :-1] * vs.maskW[2:-2, 2:-2, :-1]
                    + vs.K_33[2:-2, 2:-2, :-1]
                    * (tr[2:-2, 2:-2, 1:, vs.taup1] - tr[2:-2, 2:-2, :-1, vs.taup1])
                    / vs.dzw[npx.newaxis, npx.newaxis, :-1]
                    * vs.maskW[2:-2, 2:-2, :-1]
                ),
            )

            out["P_diss_iso"] = vs.P_diss_iso

    return KernelOutput(**out)


@veros_routine
def isoneutral_diffusion(state, tr, istemp):
    """
    Isopycnal diffusion for tracer,
    following functional formulation by Griffies et al
    Dissipation is calculated and stored in P_diss_iso
    T/S changes are added to dtemp_iso/dsalt_iso
    """
    vs = state.variables
    vs.update(isoneutral_diffusion_kernel(state, tr, istemp, iso=True))


@veros_routine
def isoneutral_skew_diffusion(state, tr, istemp):
    """
    Isopycnal skew diffusion for tracer,
    following functional formulation by Griffies et al
    Dissipation is calculated and stored in P_diss_skew
    T/S changes are added to dtemp_iso/dsalt_iso
    """
    vs = state.variables
    vs.update(isoneutral_diffusion_kernel(state, tr, istemp, iso=False))



import abc

# do not import veros.core here!
from veros import settings, time, signals, distributed, progress, runtime_settings as rs, logger
from veros.state import get_default_state
from veros.plugins import load_plugin
from veros.routines import veros_routine, is_veros_routine
from veros.timer import timer_context


class VerosSetup(metaclass=abc.ABCMeta):
    """Main class for Veros, used for building a model and running it.

    Note:
        This class is meant to be subclassed. Subclasses need to implement the
        methods :meth:`set_parameter`, :meth:`set_topography`, :meth:`set_grid`,
        :meth:`set_coriolis`, :meth:`set_initial_conditions`, :meth:`set_forcing`,
        :meth:`set_diagnostics`, and :meth:`after_timestep`.

    Example:
        >>> import matplotlib.pyplot as plt
        >>> from veros import VerosSetup
        >>>
        >>> class MyModel(VerosSetup):
        >>>     ...
        >>>
        >>> simulation = MyModel()
        >>> simulation.run()
        >>> plt.imshow(simulation.state.variables.psi[..., 0])
        >>> plt.show()

    """

    __veros_plugins__ = tuple()

    def __init__(self, override=None):
        self.override_settings = override or {}

        # this should be the first time the core routines are imported
        import veros.core  # noqa: F401

        self._plugin_interfaces = tuple(load_plugin(p) for p in self.__veros_plugins__)
        self._setup_done = False

        self.state = get_default_state(plugin_interfaces=self._plugin_interfaces)

    @abc.abstractmethod
    def set_parameter(self, state):
        """To be implemented by subclass.

        First function to be called during setup.
        Use this to modify the model settings.

        Example:
          >>> def set_parameter(self, state):
          >>>     settings = state.settings
          >>>     settings.nx, settings.ny, settings.nz = (360, 120, 50)
          >>>     settings.coord_degree = True
          >>>     settings.enable_cyclic = True
        """
        pass

    @abc.abstractmethod
    def set_initial_conditions(self, state):
        """To be implemented by subclass.

        May be used to set initial conditions.

        Example:
          >>> @veros_method
          >>> def set_initial_conditions(self, state):
          >>>     vs = state.variables
          >>>     vs.u = update(vs.u, at[:, :, :, vs.tau], npx.random.rand(vs.u.shape[:-1]))
        """
        pass

    @abc.abstractmethod
    def set_grid(self, state):
        """To be implemented by subclass.

        Has to set the grid spacings :attr:`dxt`, :attr:`dyt`, and :attr:`dzt`,
        along with the coordinates of the grid origin, :attr:`x_origin` and
        :attr:`y_origin`.

        Example:
          >>> @veros_method
          >>> def set_grid(self, state):
          >>>     vs = state.variables
          >>>     vs.x_origin, vs.y_origin = 0, 0
          >>>     vs.dxt = [0.1, 0.05, 0.025, 0.025, 0.05, 0.1]
          >>>     vs.dyt = 1.
          >>>     vs.dzt = [10, 10, 20, 50, 100, 200]
        """
        pass

    @abc.abstractmethod
    def set_coriolis(self, state):
        """To be implemented by subclass.

        Has to set the Coriolis parameter :attr:`coriolis_t` at T grid cells.

        Example:
          >>> @veros_method
          >>> def set_coriolis(self, state):
          >>>     vs = state.variables
          >>>     vs.coriolis_t = 2 * vs.omega * npx.sin(vs.yt[npx.newaxis, :] / 180. * vs.pi)
        """
        pass

    @abc.abstractmethod
    def set_topography(self, state):
        """To be implemented by subclass.

        Must specify the model topography by setting :attr:`kbot`.

        Example:
          >>> @veros_method
          >>> def set_topography(self, state):
          >>>     vs = state.variables
          >>>     vs.kbot = update(vs.kbot, at[...], 10)
          >>>     # add a rectangular island somewhere inside the domain
          >>>     vs.kbot = update(vs.kbot, at[10:20, 10:20], 0)
        """
        pass

    @abc.abstractmethod
    def set_forcing(self, state):
        """To be implemented by subclass.

        Called before every time step to update the external forcing, e.g. through
        :attr:`forc_temp_surface`, :attr:`forc_salt_surface`, :attr:`surface_taux`,
        :attr:`surface_tauy`, :attr:`forc_tke_surface`, :attr:`temp_source`, or
        :attr:`salt_source`. Use this method to implement time-dependent forcing.

        Example:
          >>> @veros_method
          >>> def set_forcing(self, state):
          >>>     vs = state.variables
          >>>     current_month = (vs.time / (31 * 24 * 60 * 60)) % 12
          >>>     vs.surface_taux = vs._windstress_data[:, :, current_month]
        """
        pass

    @abc.abstractmethod
    def set_diagnostics(self, state):
        """To be implemented by subclass.

        Called before setting up the :ref:`diagnostics <diagnostics>`. Use this method e.g. to
        mark additional :ref:`variables <variables>` for output.

        Example:
          >>> @veros_method
          >>> def set_diagnostics(self, state):
          >>>     state.diagnostics['snapshot'].output_variables += ['drho', 'dsalt', 'dtemp']
        """
        pass

    @abc.abstractmethod
    def after_timestep(self, state):
        """Called at the end of each time step. Can be used to define custom, setup-specific
        events.
        """
        pass

    def _ensure_setup_done(self):
        if not self._setup_done:
            raise RuntimeError("setup() method has to be called before running the model")

    def setup(self):
        from veros import diagnostics, restart
        from veros.core import numerics, external, isoneutral

        setup_funcs = (
            self.set_parameter,
            self.set_grid,
            self.set_coriolis,
            self.set_topography,
            self.set_initial_conditions,
            self.set_diagnostics,
            self.set_forcing,
            self.after_timestep,
        )

        for f in setup_funcs:
            if not is_veros_routine(f):
                raise RuntimeError(
                    f"{f.__name__} method is not a Veros routine. Please make sure to decorate it "
                    "with @veros_routine and try again."
                )

        logger.info("Running model setup")

        with self.state.timers["setup"]:
            with self.state.settings.unlock():
                self.set_parameter(self.state)

                for setting, value in self.override_settings.items():
                    setattr(self.state.settings, setting, value)

            settings.check_setting_conflicts(self.state.settings)
            distributed.validate_decomposition(self.state.dimensions)

            self.state.initialize_variables()

            self.state.diagnostics.update(diagnostics.create_default_diagnostics(self.state))

            for plugin in self._plugin_interfaces:
                for diagnostic in plugin.diagnostics:
                    self.state.diagnostics[diagnostic.name] = diagnostic()

            self.set_grid(self.state)
            numerics.calc_grid(self.state)

            self.set_coriolis(self.state)
            numerics.calc_beta(self.state)

            self.set_topography(self.state)
            numerics.calc_topo(self.state)

            self.set_initial_conditions(self.state)
            numerics.calc_initial_conditions(self.state)

            if self.state.settings.enable_streamfunction:
                external.streamfunction_init(self.state)

            for plugin in self._plugin_interfaces:
                plugin.setup_entrypoint(self.state)

            self.set_diagnostics(self.state)
            diagnostics.initialize(self.state)
            restart.read_restart(self.state)

            self.set_forcing(self.state)
            isoneutral.check_isoneutral_slope_crit(self.state)

        self._setup_done = True

    @veros_routine
    def step(self, state):
        from veros import diagnostics, restart
        from veros.core import idemix, eke, tke, momentum, thermodynamics, advection, utilities, isoneutral, numerics

        self._ensure_setup_done()

        vs = state.variables
        settings = state.settings

        with state.timers["diagnostics"]:
            restart.write_restart(state)

        with state.timers["main"]:
            with state.timers["forcing"]:
                self.set_forcing(state)

            if state.settings.enable_idemix:
                with state.timers["idemix"]:
                    idemix.set_idemix_parameter(state)

            with state.timers["eke"]:
                eke.set_eke_diffusivities(state)

            with state.timers["tke"]:
                tke.set_tke_diffusivities(state)

            with state.timers["momentum"]:
                momentum.momentum(state)

            with state.timers["thermodynamics"]:
                thermodynamics.thermodynamics(state)

            if settings.enable_eke or settings.enable_tke or settings.enable_idemix:
                with state.timers["advection"]:
                    advection.calculate_velocity_on_wgrid(state)

            with state.timers["eke"]:
                if state.settings.enable_eke:
                    eke.integrate_eke(state)

            with state.timers["idemix"]:
                if state.settings.enable_idemix:
                    idemix.integrate_idemix(state)

            with state.timers["tke"]:
                if state.settings.enable_tke:
                    tke.integrate_tke(state)

            with state.timers["boundary_exchange"]:
                vs.u = utilities.enforce_boundaries(vs.u, settings.enable_cyclic_x)
                vs.v = utilities.enforce_boundaries(vs.v, settings.enable_cyclic_x)
                if settings.enable_tke:
                    vs.tke = utilities.enforce_boundaries(vs.tke, settings.enable_cyclic_x)
                if settings.enable_eke:
                    vs.eke = utilities.enforce_boundaries(vs.eke, settings.enable_cyclic_x)
                if settings.enable_idemix:
                    vs.E_iw = utilities.enforce_boundaries(vs.E_iw, settings.enable_cyclic_x)

            with state.timers["momentum"]:
                momentum.vertical_velocity(state)

        with state.timers["plugins"]:
            for plugin in self._plugin_interfaces:
                with state.timers[plugin.name]:
                    plugin.run_entrypoint(state)

        vs.itt = vs.itt + 1
        vs.time = vs.time + settings.dt_tracer

        self.after_timestep(state)

        with state.timers["diagnostics"]:
            if not numerics.sanity_check(state):
                raise RuntimeError(f"solution diverged at iteration {vs.itt}")

            isoneutral.isoneutral_diag_streamfunction(state)
            diagnostics.diagnose(state)
            diagnostics.output(state)

        # NOTE: benchmarks parse this, do not change / remove
        logger.debug(" Time step took {:.2f}s", state.timers["main"].last_time)

        # permutate time indices
        vs.taum1, vs.tau, vs.taup1 = vs.tau, vs.taup1, vs.taum1

    def run(self, show_progress_bar=None):
        """Main routine of the simulation.

        Note:
            Make sure to call :meth:`setup` prior to this function.

        Arguments:
            show_progress_bar (:obj:`bool`, optional): Whether to show fancy progress bar via tqdm.
                By default, only show if stdout is a terminal and Veros is running on a single process.

        """
        from veros import restart

        self._ensure_setup_done()

        vs = self.state.variables
        settings = self.state.settings

        time_length, time_unit = time.format_time(settings.runlen)
        logger.info(f"\nStarting integration for {time_length:.1f} {time_unit}")

        start_time = vs.time

        # disable timers for first iteration
        timer_context.active = False

        pbar = progress.get_progress_bar(self.state, use_tqdm=show_progress_bar)

        try:
            with signals.signals_to_exception(), pbar:
                while vs.time - start_time < settings.runlen:
                    self.step(self.state)

                    if not timer_context.active:
                        timer_context.active = True

                    pbar.advance_time(settings.dt_tracer)

        except:  # noqa: E722
            logger.critical(f"Stopping integration at iteration {vs.itt}")
            raise

        else:
            logger.success("Integration done\n")

        finally:
            restart.write_restart(self.state, force=True)
            self._timing_summary()

    def _timing_summary(self):
        timing_summary = []

        timing_summary.extend(
            [
                "",
                "Timing summary:",
                "(excluding first iteration)",
                "---",
                " setup time               = {:.2f}s".format(self.state.timers["setup"].total_time),
                " main loop time           = {:.2f}s".format(self.state.timers["main"].total_time),
                "   forcing                = {:.2f}s".format(self.state.timers["forcing"].total_time),
                "   momentum               = {:.2f}s".format(self.state.timers["momentum"].total_time),
                "     pressure             = {:.2f}s".format(self.state.timers["pressure"].total_time),
                "     friction             = {:.2f}s".format(self.state.timers["friction"].total_time),
                "   thermodynamics         = {:.2f}s".format(self.state.timers["thermodynamics"].total_time),
            ]
        )

        if rs.profile_mode:
            timing_summary.extend(
                [
                    "     lateral mixing       = {:.2f}s".format(self.state.timers["isoneutral"].total_time),
                    "     vertical mixing      = {:.2f}s".format(self.state.timers["vmix"].total_time),
                    "     equation of state    = {:.2f}s".format(self.state.timers["eq_of_state"].total_time),
                ]
            )

        timing_summary.extend(
            [
                "   advection              = {:.2f}s".format(self.state.timers["advection"].total_time),
                "   EKE                    = {:.2f}s".format(self.state.timers["eke"].total_time),
                "   IDEMIX                 = {:.2f}s".format(self.state.timers["idemix"].total_time),
                "   TKE                    = {:.2f}s".format(self.state.timers["tke"].total_time),
                "   boundary exchange      = {:.2f}s".format(self.state.timers["boundary_exchange"].total_time),
                " diagnostics and I/O      = {:.2f}s".format(self.state.timers["diagnostics"].total_time),
                " plugins                  = {:.2f}s".format(self.state.timers["plugins"].total_time),
            ]
        )

        timing_summary.extend(
            [
                "   {:<22} = {:.2f}s".format(plugin.name, self.state.timers[plugin.name].total_time)
                for plugin in self._plugin_interfaces
            ]
        )

        logger.debug("\n".join(timing_summary))

        if rs.profile_mode:
            print_profile_summary(self.state.profile_timers, self.state.timers["main"].total_time)


def print_profile_summary(profile_timers, main_loop_time):
    profile_timings = ["", "Profile timings:", "[total time spent (% of main loop)]", "---"]
    maxwidth = max(len(k) for k in profile_timers.keys())
    profile_format_string = "{{:<{}}} = {{:.2f}}s ({{:.2f}}%)".format(maxwidth)
    main_loop_time = max(main_loop_time, 1e-8)  # prevent division by 0

    for name, timer in profile_timers.items():
        this_time = timer.total_time
        if this_time == 0:
            continue

        profile_timings.append(profile_format_string.format(name, this_time, 100 * this_time / main_loop_time))

    logger.diagnostic("\n".join(profile_timings))
